// this :
#include <BatchLab/Core/BaseFitter.h>

#include <AIDA/ITree.h>
#include <AIDA/IHistogram1D.h>
#include <AIDA/IProfile1D.h>
#include <AIDA/IAxis.h>
#include <AIDA/IFunctionFactory.h>
#include <AIDA/IFunction.h>
#include <AIDA/IDataPointSet.h>
#include <AIDA/IDataPoint.h>
#include <AIDA/IMeasurement.h>
#include <AIDA/IAnnotation.h>
#include <AIDA/ICloud1D.h>
#include <AIDA/ICloud2D.h>

#include <Slash/Core/IWriter.h>
#include <Lib/Out.h>
#include <Lib/smanip.h>
#include <Lib/sout.h>
#include <Lib/Cast.h>
#include <Lib/Debug.h>

#include <BatchLab/Core/FitResult.h>

#include <BatchLab/Core/Histogram.h> //FIXME

#include <math.h>

//////////////////////////////////////////////////////////////////////////////
void BatchLab::BaseFitter::setFitInfos(
 AIDA::IAnnotation& aAnnotation
,AIDA::IFitResult& aResult
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  aAnnotation.removeItem("fit_quality");
  aAnnotation.addItem("fit_quality",Lib::smanip::tostring(aResult.quality()));

  aAnnotation.removeItem("fit_ndf");
  aAnnotation.addItem("fit_ndf",Lib::smanip::tostring(aResult.ndf()));

  std::vector<std::string> names = aResult.fittedParameterNames();
  std::vector<double> values = aResult.fittedParameters();
  std::vector<double> errs = aResult.errors();
  if(names.size()!=values.size()) return;
  if(names.size()!=errs.size()) return;

  unsigned int namen = names.size();
  for(unsigned index=0;index<namen;index++){
    std::string key = "fit_param_"+names[index];
    aAnnotation.removeItem(key);
    aAnnotation.addItem(key,
      Lib::smanip::tostring(values[index])+" "+
      Lib::smanip::tostring(errs[index]));
  }
}

//////////////////////////////////////////////////////////////////////////////
BatchLab::BaseFitter::BaseFitter(
 Slash::Core::IWriter& aPrinter
,AIDA::IFunctionFactory& aFunctionFactory
,const std::string& aEngine
)
:fPrinter(aPrinter)
,fFunctionFactory(aFunctionFactory)
,fEngineName(aEngine)
,fCounter(0)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Debug::increment("BatchLab::BaseFitter");
}
//////////////////////////////////////////////////////////////////////////////
BatchLab::BaseFitter::~BaseFitter(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Debug::decrement("BatchLab::BaseFitter");
}
//////////////////////////////////////////////////////////////////////////////
void* BatchLab::BaseFitter::cast(
 const std::string& aClass
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if_Lib_SCast(AIDA::IFitter)
  else if_Lib_SCast(BatchLab::BaseFitter)
  else return 0;
}

//////////////////////////////////////////////////////////////////////////////
template <class T> static AIDA::IFitResult* BatchLab_BaseFitter_fit(
 BatchLab::BaseFitter& aFitter
,Slash::Core::IWriter& aPrinter
,T& aHistogram
,AIDA::IFunction& aFunction
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aHistogram.entries()==0) {
    Lib::Out out(aPrinter);
    out << "BatchLab::BaseFitter::fit :"
        << " histogram has no entries." << Lib::endl;
    return 0;
  }

  int nbin = aHistogram.axis().bins();
  if(nbin<=0) {
    Lib::Out out(aPrinter);
    out << "BatchLab::BaseFitter::fit :"
        << " has no bins." << Lib::endl;
    return 0;
  }

  int number = 0;
 {for(int count=0;count<nbin;count++) { 
    int nent = aHistogram.binEntries(count);
    if(nent==0) continue;  // Do not take bin without entry.
    number++;
  }}

  std::vector<double> xs(number);
  std::vector<double> vs(number);
  std::vector<double> es(number);
  int bini = 0;
  for(int index=0;index<nbin;index++) { 
    int nent = aHistogram.binEntries(index);
    if(nent==0) continue;  // Do not take bin without entry.
    // Take middle of bin.
    double xx = 
      (aHistogram.axis().binLowerEdge(index)+
       aHistogram.axis().binUpperEdge(index))/2.;
    double val = aHistogram.binHeight(index);
    double err = ::sqrt(::fabs(val)); // Take statistical error.
    if(err==0.) { 
      Lib::Out out(aPrinter);
      out << "BatchLab::BaseFitter::fit :"
          << " null error for bin " << index 
          << " x " << xx << " value " << val
          << ". Set error to 0.01."
          << Lib::endl;
      //return 0;
      err = 0.01; //FIXME
    }
    xs[bini] = xx;
    vs[bini] = val;
    es[bini] = err;
    bini++;
  }
  if(xs.size()==0) {
    Lib::Out out(aPrinter);
    out << "BatchLab::BaseFitter::fit :"
        << " no starting points." << Lib::endl;
    return 0;
  }

  double chi2;
  std::vector<double> errs;
  if(!aFitter.baseFit(aFunction,xs,vs,es,chi2,errs)) return 0;

  const std::vector<double>& func_params = aFunction.parameters();
  int degreeOfFreedom = xs.size()-func_params.size();
  double quality = degreeOfFreedom?(chi2/degreeOfFreedom):0;

  BatchLab::FitResult* result = new BatchLab::FitResult
    (aFunction,errs,degreeOfFreedom,quality);

  BatchLab::BaseFitter::setFitInfos(aHistogram.annotation(),*result);

  return result;
}

//////////////////////////////////////////////////////////////////////////////
static AIDA::IFitResult* BatchLab_BaseFitter_fit(
 BatchLab::BaseFitter& aFitter
,Slash::Core::IWriter& aPrinter
,AIDA::ICloud1D& aCloud
,AIDA::IFunction& aFunction
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aCloud.isConverted()) {

    //break const :
    AIDA::IHistogram1D* histo = (AIDA::IHistogram1D*)&(aCloud.histogram());
    return BatchLab_BaseFitter_fit(aFitter,aPrinter,*histo,aFunction);
    
  } else {

    int number = aCloud.entries();
    if(number==0) {
      Lib::Out out(aPrinter);
      out << "BatchLab::BaseFitter::fit(Cloud1D) :"
          << " cloud has no entries." << Lib::endl;
      return 0;
    }

    std::vector<double> xs(number);
    std::vector<double> vs(number);
    std::vector<double> es(number);

    for(unsigned int index=0;index<number;index++) { 
      double xx = aCloud.value(index);
      double val = aCloud.weight(index);
      double err = ::sqrt(::fabs(val)); // Take statistical error.
      if(err==0.) { 
        Lib::Out out(aPrinter);
        out << "BatchLab::BaseFitter::fit(Cloud1D) :"
            << " null error for point " << index
            << " x " << xx << " value " << val
            << ". Set error to 0.01."
            << Lib::endl;
        //return 0;
        err = 0.01; //FIXME
      }
      es[index] = xx;
      vs[index] = val;
      es[index] = err;
    }

    double chi2;
    std::vector<double> errs;
    if(!aFitter.baseFit(aFunction,xs,vs,es,chi2,errs)) return 0;

    const std::vector<double>& func_params = aFunction.parameters();
    int degreeOfFreedom = xs.size()-func_params.size();
    double quality = degreeOfFreedom?(chi2/degreeOfFreedom):0;

    BatchLab::FitResult* result = new BatchLab::FitResult
      (aFunction,errs,degreeOfFreedom,quality);

    BatchLab::BaseFitter::setFitInfos(aCloud.annotation(),*result);

    return result;
  }
}
//////////////////////////////////////////////////////////////////////////////
static AIDA::IFitResult* BatchLab_BaseFitter_fit(
 BatchLab::BaseFitter& aFitter
,Slash::Core::IWriter& aPrinter
,AIDA::ICloud2D& aCloud
,AIDA::IFunction& aFunction
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aCloud.isConverted()) {

    //return BatchLab_BaseFitter_fit
    //         (aFitter,aPrinter,aCloud.histogram(),aFunction);
    Lib::Out out(aPrinter);
    out << "BatchLab::BaseFitter::fit(Cloud2D) :"
        << " don't know how to fit an histogram 2D." << Lib::endl;
    return 0;

  } else {

    int number = aCloud.entries();
    if(number==0) {
      Lib::Out out(aPrinter);
      out << "BatchLab::BaseFitter::fit(Cloud2D) :"
          << " cloud has no entries." << Lib::endl;
      return 0;
    }

    std::vector<double> xs(number);
    std::vector<double> vs(number);
    std::vector<double> es(number);

    for(unsigned int index=0;index<number;index++) { 
      double xx = aCloud.valueX(index);
      double val = aCloud.valueY(index);
      double err = ::sqrt(::fabs(val)); // Take statistical error.
      if(err==0.) { 
        Lib::Out out(aPrinter);
        out << "BatchLab::BaseFitter::fit(Cloud2D) :"
            << " null error for point " << index
            << " x " << xx << " value " << val
            << ". Set error to 0.01."
            << Lib::endl;
        //return 0;
        err = 0.01; //FIXME
      }
      xs[index] = xx;
      vs[index] = val;
      es[index] = err;
    }

    double chi2;
    std::vector<double> errs;
    if(!aFitter.baseFit(aFunction,xs,vs,es,chi2,errs)) return 0;

    const std::vector<double>& func_params = aFunction.parameters();
    int degreeOfFreedom = xs.size()-func_params.size();
    double quality = degreeOfFreedom?(chi2/degreeOfFreedom):0;

    BatchLab::FitResult* result = new BatchLab::FitResult
      (aFunction,errs,degreeOfFreedom,quality);

    BatchLab::BaseFitter::setFitInfos(aCloud.annotation(),*result);

    return result;
  }
}

//////////////////////////////////////////////////////////////////////////////
AIDA::IFitResult* BatchLab::BaseFitter::fit(
 AIDA::IBaseHistogram& aHistogram
,AIDA::IFunction& aFunction
#if (AIDA_VERSION_MAJOR<=3) && (AIDA_VERSION_MINOR<=2) 
#else
,const std::string& //aRange
#endif
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(AIDA::IHistogram1D* histo = Lib_Cast(aHistogram,AIDA::IHistogram1D)) {
    return BatchLab_BaseFitter_fit(*this,fPrinter,*histo,aFunction);
  } else if(AIDA::IProfile1D* prof = Lib_Cast(aHistogram,AIDA::IProfile1D)) {
    return BatchLab_BaseFitter_fit(*this,fPrinter,*prof,aFunction);
  } else if(AIDA::ICloud1D* cloud = Lib_Cast(aHistogram,AIDA::ICloud1D)) {
    return BatchLab_BaseFitter_fit(*this,fPrinter,*cloud,aFunction);
  } else if(AIDA::ICloud2D* cloud = Lib_Cast(aHistogram,AIDA::ICloud2D)) {
    return BatchLab_BaseFitter_fit(*this,fPrinter,*cloud,aFunction);
  } else {
    Lib::Out out(fPrinter);
    out << "BatchLab::BaseFitter::fit :"
        << " base histogram " << Lib::sout(aHistogram.title())
        << " is not an AIDA::IHistogram1D or an AIDA::IProfile1D." 
        << Lib::endl;
    return 0;
  }
}

//////////////////////////////////////////////////////////////////////////////
template <class T> static bool HalfLowerMeanPoint (
 T& aHisto
,double& aX
,double& aV
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  aX = 0;
  aV = 0;
  int binn = aHisto.axis().bins();
  int end = binn/2;
  int n = 0;
  double x = 0;
  double v = 0;
  for(int count=0;count<end;count++) { 
    if(aHisto.binEntries(count)==0) continue;  // Do not take empty bins.
    // Take middle of Bin.
    x += 
      (aHisto.axis().binLowerEdge(count)+aHisto.axis().binUpperEdge(count))/2.;
    v += aHisto.binHeight(count);
    n++;
  }
  if(n<=0) return false;
  aX = x/n;
  aV = v/n;
  return true;
}
//////////////////////////////////////////////////////////////////////////////
template <class T> static bool HalfUpperMeanPoint(
 T& aHisto
,double& aX
,double& aV
)
//////////////////////////////////////////////////////////////////////////////
//  Used to find an end point for linear, exponential fit.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  aX = 0.;
  aV = 0.;
  int binn = aHisto.axis().bins();
  int begin = binn/2;
  int n = 0;
  double x = 0.;
  double v = 0.;
  for(int count=begin;count<binn;count++) {
    if(aHisto.binEntries(count)==0) continue;  // Do not take empty bins.
    // Take middle of Bin.
    x += 
      (aHisto.axis().binUpperEdge(count)+aHisto.axis().binLowerEdge(count))/2.;
    v += aHisto.binHeight(count);
    n++;
  }
  if(n<=0) return false;
  aX = x/n;
  aV = v/n;
  return true;
}

//////////////////////////////////////////////////////////////////////////////
template <class T> static AIDA::IFitResult* BatchLab_BaseFitter_fit(
 BatchLab::BaseFitter& aFitter
,Slash::Core::IWriter& aPrinter
,AIDA::IFunctionFactory& aFunctionFactory
,int& aCounter
,T& aHistogram
,const std::string& aModel
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  // Build a name :
  std::string name;
  Lib::smanip::printf(name,128,"fitHistogram_%d",aCounter);
  aCounter++;
  std::string title = name; //FIXME
  std::string model = aModel;
  Lib::smanip::touppercase(model);
  int degree;

  if( (model=="G") || (model=="GAUSSIAN") ) {

    double h = aHistogram.maxBinHeight() - aHistogram.minBinHeight();
    double m = aHistogram.mean();
    double w = aHistogram.rms();

    AIDA::IFunction* function = 
      aFunctionFactory.createFunctionByName(name,"G");
    if(!function) return 0;
    std::vector<double> ps(3);
    ps[0] = h;
    ps[1] = m;
    ps[2] = w;
    function->setParameters(ps);

    return BatchLab_BaseFitter_fit(aFitter,aPrinter,aHistogram,*function);

  } else if( (model=="BW") || (model=="BREITWIGNER") ) {

    double h = aHistogram.maxBinHeight() - aHistogram.minBinHeight();
    double m = aHistogram.mean();
    double w = aHistogram.rms();

    AIDA::IFunction* function = 
      aFunctionFactory.createFunctionByName(name,"BW");
    if(!function) return 0;
    std::vector<double> ps(3);
    ps[0] = h;
    ps[1] = m;
    ps[2] = w;
    function->setParameters(ps);

    return BatchLab_BaseFitter_fit(aFitter,aPrinter,aHistogram,*function);

  } else if( (model=="E") || (model=="EXPONENTIAL") ) {

    double xx1,xx2;
    double val1,val2;
    bool status1 = HalfLowerMeanPoint(aHistogram,xx1,val1);
    bool status2 = HalfUpperMeanPoint(aHistogram,xx2,val2);
    if( !status1 || !status2 || ((xx2-xx1)==0.) ) {
      Lib::Out out(aPrinter);
      out << "BatchLab::BaseFitter::fit : can't compute starting values." 
          << Lib::endl;
      return 0;
    }
    
    // Deduce p0,p1 of p0*exp(p1*x) with the two upper points :
    double p1 = ::log(val2/val1) / ( xx2 - xx1);
    double p0 = val1 / exp ( p1 * xx1 );

    AIDA::IFunction* function = 
      aFunctionFactory.createFunctionByName(name,"E");
    if(!function) return 0;
    std::vector<double> ps(2);
    ps[0] = p0;
    ps[1] = p1;
    function->setParameters(ps);

    return BatchLab_BaseFitter_fit(aFitter,aPrinter,aHistogram,*function);

  } else if(model=="EHBOOK") {

    double xx1,xx2;
    double val1,val2;
    bool status1 = HalfLowerMeanPoint(aHistogram,xx1,val1);
    bool status2 = HalfUpperMeanPoint(aHistogram,xx2,val2);
    if( !status1 || !status2 || ((xx2-xx1)==0.) ) {
      Lib::Out out(aPrinter);
      out << "BatchLab::BaseFitter::fit : can't compute starting values." 
          << Lib::endl;
      return 0;
    }
    
    // Deduce a,b of exp(a+bx) with the two upper points :
    double b = ::log(val2/val1) / ( xx2 - xx1);
    double a = ::log( val1 / exp ( b * xx1 ) );

    AIDA::IFunction* function = 
      aFunctionFactory.createFunctionByName(name,"EHBOOK");
    if(!function) return 0;
    std::vector<double> ps(2);
    ps[0] = a;
    ps[1] = b;
    function->setParameters(ps);

    return BatchLab_BaseFitter_fit(aFitter,aPrinter,aHistogram,*function);

  } else if(Lib::smanip::polynomialDegree(model,degree)) {

    double h = aHistogram.maxBinHeight() - aHistogram.minBinHeight();
    double m = aHistogram.mean();
    double w = aHistogram.rms();

    AIDA::IFunction* function = 
      aFunctionFactory.createFunctionByName(name,model);
    if(!function) return 0;
    std::vector<double> ps;
    ps.resize(degree+1,1);
    if(degree==2) {
      ps[0] = h;
      ps[1] = m;
      ps[2] = w;
    } 
    function->setParameters(ps);

    return BatchLab_BaseFitter_fit(aFitter,aPrinter,aHistogram,*function);

  } 

  Lib::Out out(aPrinter);
  out << "BatchLab::BaseFitter::fit :"
      << " unknown fit model " << Lib::sout(model)
      << Lib::endl;

  return 0;
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IFitResult* BatchLab::BaseFitter::fit(
 AIDA::IBaseHistogram& aHistogram
,const std::string& aModel
#if (AIDA_VERSION_MAJOR<=3) && (AIDA_VERSION_MINOR<=2) 
#else
,const std::string& //aRange
#endif
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(AIDA::IHistogram1D* histo = Lib_Cast(aHistogram,AIDA::IHistogram1D)) {
    return BatchLab_BaseFitter_fit
             (*this,fPrinter,fFunctionFactory,fCounter,*histo,aModel);
  } else if(AIDA::IProfile1D* prof = Lib_Cast(aHistogram,AIDA::IProfile1D)) {
    return BatchLab_BaseFitter_fit
             (*this,fPrinter,fFunctionFactory,fCounter,*prof,aModel);
/*
  } else if(AIDA::ICloud1D* cloud = Lib_Cast(aHistogram,AIDA::ICloud1D)) {
    //FIXME : is it what expected when fitting a cloud ?
    double mn = cloud->lowerEdge();
    double mx = cloud->upperEdge();
    BatchLab::Histogram1D h(fPrinter,"","",100,mn,mx);
    cloud->fillHistogram(h);
    return BatchLab_BaseFitter_fit
             (*this,fPrinter,fFunctionFactory,fCounter,h,aModel);
*/
  } else {
    Lib::Out out(fPrinter);
    out << "BatchLab::BaseFitter::fit(model) :"
        << " base histogram " << Lib::sout(aHistogram.title())
        << " is not an AIDA::IHistogram1D or an AIDA::IProfile1D." 
        << Lib::endl;
    return 0;
  }
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IFitResult* BatchLab::BaseFitter::fit(
 AIDA::IDataPointSet& aDataPointSet
,AIDA::IFunction& aFunction
#if (AIDA_VERSION_MAJOR<=3) && (AIDA_VERSION_MINOR<=2) 
#else
,const std::string& //aRange
#endif
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  int dim = aDataPointSet.dimension();
  if(dim!=2) {
    Lib::Out out(fPrinter);
    out << "BatchLab::BaseFitter::fit :"
        << " dataPointSet " << Lib::sout(aDataPointSet.title())
        << " is not of dimension 2." 
        << Lib::endl;
    return 0;
  }
  int number = aDataPointSet.size();
  if(number<=0) return 0;

  std::vector<double> xs(number);
  std::vector<double> vs(number);
  std::vector<double> es(number);
  for(int index=0;index<number;index++) {
    double x = aDataPointSet.point(index)->coordinate(0)->value();
    double y = aDataPointSet.point(index)->coordinate(1)->value();
    //double e = aDataPointSet.point(index)->coordinate(2)->value();
    //double e = 0;
    double e = ::sqrt(::fabs(y));
    if(e==0) e = 0.01; //WARNING : must not be 0.
    xs[index] = x;
    vs[index] = y;
    es[index] = e;
  }
  
  double chi2;
  std::vector<double> errs;
  if(!baseFit(aFunction,xs,vs,es,chi2,errs)) return 0;

  const std::vector<double>& func_params = aFunction.parameters();
  int degreeOfFreedom = xs.size()-func_params.size();
  double quality = degreeOfFreedom?(chi2/degreeOfFreedom):0;

  FitResult* result = new FitResult(aFunction,errs,degreeOfFreedom,quality);

  BatchLab::BaseFitter::setFitInfos(aDataPointSet.annotation(),*result);

  return result;
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IFitResult* BatchLab::BaseFitter::fit(
 AIDA::IDataPointSet& aDataPointSet
,const std::string& aModel
#if (AIDA_VERSION_MAJOR<=3) && (AIDA_VERSION_MINOR<=2) 
#else
,const std::string& //aRange
#endif
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  int dim = aDataPointSet.dimension();
  if(dim!=2) {
    Lib::Out out(fPrinter);
    out << "BatchLab::BaseFitter::fit :"
        << " dataPointSet " << Lib::sout(aDataPointSet.title())
        << " is not of dimension 2." 
        << Lib::endl;
    return 0;
  }

  // Build a name :
  std::string name;
  Lib::smanip::printf(name,128,"fitDataPointSet_%d",fCounter);
  fCounter++;
  std::string title = name; //FIXME
  std::string model = aModel;
  Lib::smanip::touppercase(model);
  int degree;

  if( (model=="G") || (model=="GAUSSIAN") ) {

    double h = 1; //FIXME
    double m = 0;
    double w = 0;

    AIDA::IFunction* function = 
      fFunctionFactory.createFunctionByName(name,"G");
    if(!function) return 0;
    std::vector<double> ps(3);
    ps[0] = h;
    ps[1] = m;
    ps[2] = w;
    function->setParameters(ps);

    return fit(aDataPointSet,*function);

  } else if( (model=="BW") || (model=="BREITWIGNER") ) {

    double h = 1; //FIXME
    double m = 0;
    double w = 0;

    AIDA::IFunction* function = 
      fFunctionFactory.createFunctionByName(name,"BW");
    if(!function) return 0;
    std::vector<double> ps(3);
    ps[0] = h;
    ps[1] = m;
    ps[2] = w;
    function->setParameters(ps);

    return fit(aDataPointSet,*function);

  } else if( (model=="E") || (model=="EXPONENTIAL") ) {

    double p0 = 1;
    double p1 = 1;

    AIDA::IFunction* function = 
      fFunctionFactory.createFunctionByName(name,"E");
    if(!function) return 0;
    std::vector<double> ps(2);
    ps[0] = p0;
    ps[1] = p1;
    function->setParameters(ps);

    return fit(aDataPointSet,*function);

  } else if(model=="EHBOOK") {

    double a = 0;
    double b = 1;

    AIDA::IFunction* function = 
      fFunctionFactory.createFunctionByName(name,"EHBOOK");
    if(!function) return 0;
    std::vector<double> ps(2);
    ps[0] = a;
    ps[1] = b;
    function->setParameters(ps);

    return fit(aDataPointSet,*function);

  } else if(Lib::smanip::polynomialDegree(model,degree)) {

    AIDA::IFunction* function = 
      fFunctionFactory.createFunctionByName(name,model);
    if(!function) return 0;
    std::vector<double> ps;
    ps.resize(degree+1,1);
    function->setParameters(ps);
    return fit(aDataPointSet,*function);

  } 

  Lib::Out out(fPrinter);
  out << "BatchLab::BaseFitter::fit : unknown fit model " << Lib::sout(model)
      << Lib::endl;

  return 0;
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IFitResult* BatchLab::BaseFitter::fit(
 AIDA::IBaseHistogram& aHistogram
,const std::string& aModel
,const std::vector<double>& aParams
#if (AIDA_VERSION_MAJOR<=3) && (AIDA_VERSION_MINOR<=2) 
#else
,const std::string& //aRange
#endif
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  AIDA::IHistogram1D* histo = Lib_Cast(aHistogram,AIDA::IHistogram1D);
  if(!histo) {
    Lib::Out out(fPrinter);
    out << "BatchLab::BaseFitter::fit :"
        << " histogram " << Lib::sout(aHistogram.title())
        << " is not an AIDA::IHistogram1D." 
        << Lib::endl;
    return 0;
  }

  // Build a name :
  std::string name;
  Lib::smanip::printf(name,128,"fitHistogram_%d",fCounter);
  fCounter++;
  std::string title = name; //FIXME
  std::string model = aModel;
  Lib::smanip::touppercase(model);
  int degree;

  if( (model=="G") || (model=="GAUSSIAN") ) {

    if(aParams.size()!=3) {
      Lib::Out out(fPrinter);
      out << "BatchLab::BaseFitter::fit :"
          << " bad number of parameters (3 expected)."
          << Lib::endl;
      return 0;
    }

    AIDA::IFunction* function = 
      fFunctionFactory.createFunctionByName(name,"G");
    if(!function) return 0;
    function->setParameters(aParams);

    return fit(aHistogram,*function);

  } else if( (model=="BW") || (model=="BREITWIGNER") ) {

    if(aParams.size()!=3) {
      Lib::Out out(fPrinter);
      out << "BatchLab::BaseFitter::fit :"
          << " bad number of parameters (3 expected)"
          << Lib::endl;
      return 0;
    }

    AIDA::IFunction* function = 
      fFunctionFactory.createFunctionByName(name,"BW");
    if(!function) return 0;
    function->setParameters(aParams);

    return fit(aHistogram,*function);

  } else if( (model=="E") || (model=="EXPONENTIAL") ) {

    if(aParams.size()!=2) {
      Lib::Out out(fPrinter);
      out << "BatchLab::BaseFitter::fit :"
          << " bad number of parameters (2 expected)"
          << Lib::endl;
      return 0;
    }

    AIDA::IFunction* function = 
      fFunctionFactory.createFunctionByName(name,"E");
    if(!function) return 0;
    function->setParameters(aParams);

    return fit(aHistogram,*function);

  } else if(model=="EHBOOK") {

    if(aParams.size()!=2) {
      Lib::Out out(fPrinter);
      out << "BatchLab::BaseFitter::fit :"
          << " bad number of parameters (2 expected)"
          << Lib::endl;
      return 0;
    }

    AIDA::IFunction* function = 
      fFunctionFactory.createFunctionByName(name,"EBOOK");
    if(!function) return 0;
    function->setParameters(aParams);

    return fit(aHistogram,*function);

  } else if(Lib::smanip::polynomialDegree(model,degree)) {

    if((int)aParams.size()!=(degree+1)) {
      Lib::Out out(fPrinter);
      out << "BatchLab::BaseFitter::fit :"
          << " bad number of parameters (" << (degree+1) << " expected)"
          << Lib::endl;
      return 0;
    }

    AIDA::IFunction* function = 
      fFunctionFactory.createFunctionByName(name,model);
    if(!function) return 0;
    function->setParameters(aParams);

    return fit(aHistogram,*function);

  } else {
    Lib::Out out(fPrinter);
    out << "BatchLab::BaseFitter::fit :"
        << " unknown fit model " << Lib::sout(model)
        << Lib::endl;
    return 0;
  }
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IFitResult* BatchLab::BaseFitter::fit(
 AIDA::IDataPointSet& aDataPointSet
,const std::string& aModel
,const std::vector<double>& aParams
#if (AIDA_VERSION_MAJOR<=3) && (AIDA_VERSION_MINOR<=2) 
#else
,const std::string& //aRange
#endif
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  int dim = aDataPointSet.dimension();
  if(dim!=2) {
    Lib::Out out(fPrinter);
    out << "BatchLab::BaseFitter::fit :"
        << " dataPointSet " << Lib::sout(aDataPointSet.title())
        << " is not of dimension 2." 
        << Lib::endl;
    return 0;
  }

  // Build a name :
  std::string name;
  Lib::smanip::printf(name,128,"fitDataPointSet_%d",fCounter);
  fCounter++;
  std::string title = name; //FIXME
  std::string model = aModel;
  Lib::smanip::touppercase(model);
  int degree;

  if( (model=="G") || (model=="GAUSSIAN") ) {

    if(aParams.size()!=3) {
      Lib::Out out(fPrinter);
      out << "BatchLab::BaseFitter::fit :"
          << " bad number of parameters (3 expected)"
          << Lib::endl;
      return 0;
    }

    AIDA::IFunction* function = 
      fFunctionFactory.createFunctionByName(name,"G");
    if(!function) return 0;
    function->setParameters(aParams);

    return fit(aDataPointSet,*function);

  } else if( (model=="BW") || (model=="BREITWIGNER") ) {

    if(aParams.size()!=3) {
      Lib::Out out(fPrinter);
      out << "BatchLab::BaseFitter::fit :"
          << " bad number of parameters (3 expected)"
          << Lib::endl;
      return 0;
    }

    AIDA::IFunction* function = 
      fFunctionFactory.createFunctionByName(name,"BW");
    if(!function) return 0;
    function->setParameters(aParams);

    return fit(aDataPointSet,*function);

  } else if( (model=="E") || (model=="EXPONENTIAL") ) {

    if(aParams.size()!=2) {
      Lib::Out out(fPrinter);
      out << "BatchLab::BaseFitter::fit :"
          << " bad number of parameters (2 expected)"
          << Lib::endl;
      return 0;
    }

    AIDA::IFunction* function = 
      fFunctionFactory.createFunctionByName(name,"E");
    if(!function) return 0;
    function->setParameters(aParams);

    return fit(aDataPointSet,*function);

  } else if(model=="EHBOOK") {

    if(aParams.size()!=2) {
      Lib::Out out(fPrinter);
      out << "BatchLab::BaseFitter::fit :"
          << " bad number of parameters (2 expected)"
          << Lib::endl;
      return 0;
    }

    AIDA::IFunction* function = 
      fFunctionFactory.createFunctionByName(name,"EHBOOK");
    if(!function) return 0;
    function->setParameters(aParams);

    return fit(aDataPointSet,*function);

  } else if(Lib::smanip::polynomialDegree(model,degree)) {

    if((int)aParams.size()!=(degree+1)) {
      Lib::Out out(fPrinter);
      out << "BatchLab::BaseFitter::fit :"
          << " bad number of parameters (" << (degree+1) << " expected)"
          << Lib::endl;
      return 0;
    }

    AIDA::IFunction* function = 
      fFunctionFactory.createFunctionByName(name,model);
    if(!function) return 0;
    function->setParameters(aParams);
    return fit(aDataPointSet,*function);

  } else {
    Lib::Out out(fPrinter);
    out << "BatchLab::BaseFitter::fit : unknown fit model " << Lib::sout(model)
        << Lib::endl;
    return 0;
  }
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IFitParameterSettings& BatchLab::BaseFitter::fitParameterSettings(
 const std::string& aName
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  unsigned int number = fFitParameterSettings.size();
  for(unsigned int index=0;index<number;index++) {
    if(fFitParameterSettings[index].name()==aName) {
      return fFitParameterSettings[index];
    }
  }
  // Not found, create one :
  fFitParameterSettings.push_back(FitParameterSettings(aName));
  return fFitParameterSettings[fFitParameterSettings.size()-1];
}
//////////////////////////////////////////////////////////////////////////////
std::vector<std::string> BatchLab::BaseFitter::listParameterSettings(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  unsigned int number = fFitParameterSettings.size();
  std::vector<std::string> names(number);
  for(unsigned int index=0;index<number;index++) {
    names[index] = fFitParameterSettings[index].name();
  }
  return names;
}
//////////////////////////////////////////////////////////////////////////////
void BatchLab::BaseFitter::resetParameterSettings(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  fParameterSettings.clear(); //FIXME : is it what AIDA semanticly wants ?
}
//////////////////////////////////////////////////////////////////////////////
std::string BatchLab::BaseFitter::engineName(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fEngineName;
}

bool BatchLab::BaseFitter::setEngine(
 const std::string& /*name*/
) 
{
  return false;
}
bool BatchLab::BaseFitter::setConstraint(
 const std::string& /*expr*/
) 
{
  return false;
}
std::vector<std::string> BatchLab::BaseFitter::constraints(
) 
{
  return fConstraints;
}
void BatchLab::BaseFitter::resetConstraints(
) 
{
}
bool BatchLab::BaseFitter::setFitMethod(
 const std::string& /*name*/
) 
{
  return false;
}
std::string BatchLab::BaseFitter::fitMethodName(
) 
{
  return fFitMethodName;
}
AIDA::IDataPointSet* BatchLab::BaseFitter::createScan1D(
 AIDA::IFitData& /*d*/
,AIDA::IFunction& /*f*/
,const std::string& /*par*/
,int /*npts*/
,double /*pmin*/
,double /*pmax*/
) 
{
  return 0;
}
AIDA::IDataPointSet* BatchLab::BaseFitter::createContour(
 AIDA::IFitData& /*d*/
,AIDA::IFitResult& /*r*/
,const std::string& /*par1*/
,const std::string& /*par2*/
,int /*npts*/
,double /*up*/
) 
{
  return 0;
}
#if (AIDA_VERSION_MAJOR<=3) && (AIDA_VERSION_MINOR<=2) 
void BatchLab::BaseFitter::setUseFunctionGradient(bool /*useGrad*/) 
{
}
#else
bool BatchLab::BaseFitter::setUseFunctionGradient(bool /*useGrad*/) 
{
  return false;
}
#endif
bool BatchLab::BaseFitter::useFunctionGradient(
) 
{
  return false;
}

//FIXME
#if (AIDA_VERSION_MAJOR<=3) && (AIDA_VERSION_MINOR<=2) 
AIDA::IFitResult* BatchLab::BaseFitter::fit(
 AIDA::IFitData& /*fitData*/
,AIDA::IFunction& /*func*/
) 
{
  return 0;
}
AIDA::IFitResult* BatchLab::BaseFitter::fit(
 AIDA::IFitData& /*d*/
,const std::string& /*model*/
) 
{
  return 0;
}
AIDA::IFitResult* BatchLab::BaseFitter::fit(
 AIDA::IFitData& /*d*/
,const std::string& /*model*/
,const std::vector<double>& /*initialParameters*/
) 
{
  return 0;
}
#else
AIDA::IFitResult* BatchLab::BaseFitter::fit(
 AIDA::IFitData& //fitData
,AIDA::IFunction& //func
,const std::string& //range
) {
  return 0;
}
AIDA::IFitResult* BatchLab::BaseFitter::fit(
 AIDA::IFitData& //fitData
,const std::string& //model
,const std::string& //range
) {
  return 0;
}
AIDA::IFitResult* BatchLab::BaseFitter::fit(
 AIDA::IFitData& //fitData
,const std::string& //model
,const std::vector<double>& //initialParams
,const std::string& //range
) {
  return 0;
}

#endif
