//
//  This program is and should be standalone so that :
//     UNIX> c++ obuild_platform.cxx 
//      DOS> cl.exe obuild_platform.cxx 
// builds at first shoot.
//
//  Applying :
//      OS> obuild_platform 
// prints on stdout the string :
//   <os>-<processor>-<compiler>.
// For example : 
//   Darwin-i386-gcc_401
//
//  It can be used too with :
//      OS> obuild_platform <expression>
//  In this case it prints on stdout "yes" or "no" 
// according the results of the evaluation of the expression
// by using the local kind of os, processor.
// The syntax of the expression is :
//   <primary>
//   not <primary>  
//   <primary> [and,or] <primary>
// With primary being :
//   <string> 
// for which <string> is compared to the os, or :
//   [os,processor][=,!=]<string>
// For example :
//   Darwin
//   not Darwin
//   not os=UNIX
//   os=UNIX and processor!=intel
//
// Then, for example on a MacIntel :
//   MacIntel> obuild_platform 'Darwin and processor=intel'
// prints yes, and on a Windows :
//   DOS> obuild_platform 'os=UNIX'
// prints no
//  
//  In case of errors, some message is expected on stderr
// and the program exits with EXIT_FAILURE (1), else
// it exists with EXIT_SUCCESS (0).
//


// If possible, below get_[os,proc,compiler]() are before 
// any include to be sure that ifdef things come only 
// from the compiler...
char* get_os() {
#if defined(__APPLE__)
  static char os[] = {"Darwin"};
#elif defined(_WIN32)
  static char os[] = {"Windows_NT"};
#elif defined(__linux)
  static char os[] = {"Linux"};
#elif defined(__alpha)
  static char os[] = {"OSF1"};
#elif defined(__CYGWIN__)
  static char os[] = {"CYGWIN"};
#else
  static char os[] = {"unknown"};
#endif
  return os;
}

char* get_proc() {
#if defined(__GNUC__)

#if defined(__ppc__)
  static char proc[] = {"ppc"};
#elif defined(__ppc64__)
  static char proc[] = {"ppc64"};
#elif defined(__i386__)
  static char proc[] = {"i386"};
#elif defined(__x86_64__)
  static char proc[] = {"x86_64"};
#else
  static char proc[] = {"unknown"};
#endif

#elif defined(_MSC_VER)

#if defined(_M_IX86)
  static char proc[] = {"ix86"};
#elif defined(_M_X64)
  static char proc[] = {"x64"};
#else
  static char proc[] = {"unknown"};
#endif

#elif defined(__alpha)
  static char proc[] = {"alpha"};

#else
  static char proc[] = {"unknown"};
#endif
  return proc;
}

#if defined(__GNUC__) || defined(_MSC_VER)
#include <stdio.h>
#endif

char* get_compiler() {
#if defined(__GNUC__)
  static char compiler[128];
  sprintf(compiler,"gcc_%d%d%d",__GNUC__,__GNUC_MINOR__,__GNUC_PATCHLEVEL__);
#elif defined(_MSC_VER)
  static char compiler[128];
  //sprintf(compiler,"cl_%d_mfc_%d",_MSC_VER,_MFC_VER);
  sprintf(compiler,"cl_%d",_MSC_VER);
#elif defined(__alpha)
  static char compiler[] = {"cxx"};
#else
  static char compiler[] = {"unknown"};
#endif
  return compiler;
}

#include <string>
#include <vector>
#include <stdlib.h>
#include <stdio.h>

// From Lib :
static std::vector<std::string> Lib_smanip_words(const std::string&,
                                                 const std::string&,
                                                 bool = false);

//////////////////////////////////////////////////////////////////////////////
static bool is_os(
 const std::string& aString
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aString=="Windows_NT") {
    return true;
  } else if(aString=="Linux") {
    return true;
  } else if(aString=="Darwin") {
    return true;
  } else if(aString=="SunOS") {
    return true;
  } else if(aString=="OSF1") {
    return true;
  } else if(aString=="CYGWIN") {
    return true;
  } else if(aString=="UNIX") {
    return true;
  } else {
    fprintf(stderr,"obuild_platform : unknown operating system \"%s\".\n",
              aString.c_str());
    return false;
  }
}
//////////////////////////////////////////////////////////////////////////////
static bool is_proc(
 const std::string& aString
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aString=="ppc") {
    return true;
  } else if(aString=="ppc64") {
    return true;
  } else if(aString=="i386") {
    return true;
  } else if(aString=="x86_64") {
    return true;
  } else if(aString=="intel") {
    return true;
  } else if(aString=="powerpc") {
    return true;
  } else {
    fprintf(stderr,"obuild_platform : unknown processor \"%s\".\n",
              aString.c_str());
    return false;
  }
}
//////////////////////////////////////////////////////////////////////////////
static bool is_UNIX(
 const std::string& aString
,bool& aResult
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!is_os(aString)) {
    aResult = false;
    return false;
  }
  if(aString=="Windows_NT") {
    aResult = false;
  } else {
    aResult = true;
  }
  return true;
}
//////////////////////////////////////////////////////////////////////////////
static bool is_intel(
 const std::string& aString
,bool& aResult
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!is_proc(aString)) {
    aResult = false;
    return false;
  }
  if(aString=="i386") {
    aResult = true;
  } else if(aString=="x86_64") {
    aResult = true;
  } else {
    aResult = false;
  }
  return true;
}
//////////////////////////////////////////////////////////////////////////////
static bool is_powerpc(
 const std::string& aString
,bool& aResult
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!is_proc(aString)) {
    aResult = false;
    return false;
  }
  if(aString=="ppc") {
    aResult = true;
  } else if(aString=="ppc64") {
    aResult = true;
  } else {
    aResult = false;
  }
  return true;
}
//////////////////////////////////////////////////////////////////////////////
static bool check_os(
 const std::string& a_os
,const std::string& aString
,bool& aResult
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aString=="UNIX") {
    return is_UNIX(a_os,aResult);
  } else if(!is_os(aString)){
    return false;
  } else {
    aResult = (aString==a_os);
    return true;
  }  
}
//////////////////////////////////////////////////////////////////////////////
static bool check_proc(
 const std::string& a_proc
,const std::string& aString
,bool& aResult
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aString=="intel") {
    return is_intel(a_proc,aResult);
  } else if(aString=="powerpc") {
    return is_powerpc(a_proc,aResult);
  } else if(!is_proc(aString)){
    return false;
  } else {
    aResult = (aString==a_proc);
    return true;
  }
}
//////////////////////////////////////////////////////////////////////////////
static bool evaluate_primary(
 const std::string& a_os
,const std::string& a_proc
,const std::string& aString
,bool& aResult
)
//////////////////////////////////////////////////////////////////////////////
// aString is :
//  <string> and then string is compared to the os.
// or :
//  [os,proc][=,!=]<string>
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string::size_type pos_neq = aString.rfind("!=");
  std::string::size_type pos_eq = aString.find('=');
  std::string left,right;
  if(pos_neq!=std::string::npos) {
    left = aString.substr(0,pos_neq);
    right = aString.substr(pos_neq+2,aString.size()-(pos_neq+2));
    if(left=="os") {
      if(a_os=="unknown") {
        fprintf(stderr,"obuild_platform : unknown operating system \"%s\".\n",
                  a_os.c_str());
        aResult = false;
        return false;
      }
      bool result;
      if(!check_os(a_os,right,result)) {
        aResult = false;
        return false;
      }
      aResult = (result?false:true);
      return true;
    } else if (left=="processor") {
      if(a_proc=="*") { //Used in obuild::Builder::build_on_platform.
        aResult = true;
        return true;
      }
      if(a_proc=="unknown") {
        fprintf(stderr,"obuild_platform : unknown processor \"%s\".\n",
                  a_proc.c_str());
        aResult = false;
        return false;
      }
      bool result;
      if(!check_proc(a_proc,right,result)){
        aResult = false;
        return false;
      }
      aResult = (result?false:true);
      return true;
    } else {
      fprintf(stderr,
       "obuild_platform : unknown keyword \"%s\". os or processor expected.\n",
       left.c_str());
      aResult = false;
      return false;
    }
  } else if(pos_eq!=std::string::npos) {
    left = aString.substr(0,pos_eq);
    right = aString.substr(pos_eq+1,aString.size()-(pos_eq+1));
    if(left=="os") {
      if(a_os=="unknown") {
        fprintf(stderr,"obuild_platform : unknown operating system \"%s\".\n",
                  a_os.c_str());
        aResult = false;
        return false;
      }
      return check_os(a_os,right,aResult);
    } else if (left=="processor") {
      if(a_proc=="*") { //Used in obuild::Builder::build_on_platform.
        aResult = true;
        return true;
      }
      if(a_proc=="unknown") {
        fprintf(stderr,"obuild_platform : unknown processor \"%s\".\n",
                  a_proc.c_str());
        aResult = false;
        return false;
      }
      return check_proc(a_proc,right,aResult);
    } else {
      fprintf(stderr,
       "obuild_platform : unknown keyword \"%s\". os or processor expected.\n",
       left.c_str());
      aResult = false;
      return false;
    }

  } else {
    if(a_os=="unknown") {
      fprintf(stderr,"obuild_platform : unknown operating system \"%s\".\n",
                a_os.c_str());
      aResult = false;
      return false;
    }
    return check_os(a_os,aString,aResult);
  }
}
//////////////////////////////////////////////////////////////////////////////
static bool evaluate_expression(
 const std::string& a_os
,const std::string& a_proc
,const std::string& aString
,bool& aResult
)
//////////////////////////////////////////////////////////////////////////////
// aString is :
//   <primary>
//   not <primary>  (for backward compatibility)
// or :
//   <primary> [and,or] <primary>
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::vector<std::string> words = Lib_smanip_words(aString," ");
  if(words.size()==1) {
    return evaluate_primary(a_os,a_proc,words[0],aResult);
  } else if( (words.size()==2) && (words[0]=="not") ) {
    // For backward compatibility :
    bool result;
    if(!evaluate_primary(a_os,a_proc,words[1],result)) {
      aResult = false;
      return false;
    }
    aResult = (result?false:true);
    return true;
  } else if(words.size()==3) {
    std::string op = words[1];
    if(op=="and") {
      bool res_left;
      if(!evaluate_primary(a_os,a_proc,words[0],res_left)) {
        aResult = false;
        return false;
      }
      bool res_right;
      if(!evaluate_primary(a_os,a_proc,words[2],res_right)) {
        aResult = false;
        return false;
      }
      aResult = res_left && res_right;
      return true;
    } else if(op=="or") {
      bool res_left;
      if(!evaluate_primary(a_os,a_proc,words[0],res_left)) {
        aResult = false;
        return false;
      }
      bool res_right;
      if(!evaluate_primary(a_os,a_proc,words[2],res_right)) {
        aResult = false;
        return false;
      }
      aResult = res_left || res_right;
      return true;
    } else {
      fprintf(stderr,
        "obuild_platform : unknown keyword \"%s\". and or or expected.\n",
        op.c_str());
      aResult = false;
      return false;
    }
  } else {
    fprintf
      (stderr,"obuild_platform : only one or three words today supported.\n");
    aResult = false;
    return false;
  }
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
int main(
 int aArgc
,char** aArgv
) 
//////////////////////////////////////////////////////////////////////////////
// aArgs[0] : program name.
// aArgs[1] : an expression to be evaluated according
//            the local kind of os, processor, compiler.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::vector<std::string> args;
 {for(int index=0;index<aArgc;index++) args.push_back(aArgv[index]);}
  
  if((args.size()!=1)&&(args.size()!=2)) {
    fprintf(stderr,"obuild_platform : zero or one arguments expected.\n");
    printf("no\n");
    return EXIT_FAILURE;
  }

  std::string os(get_os());
 {char* env = getenv("OBUILD_PLATFORM_OS");
  if(env) os = std::string(env);}

  std::string proc(get_proc());
 {char* env = getenv("OBUILD_PLATFORM_PROCESSOR");
  if(env) proc = std::string(env);}

  std::string compiler(get_compiler());
 {char* env = getenv("OBUILD_PLATFORM_COMPILER");
  if(env) compiler = std::string(env);}

  if(args.size()==1) {
    printf("%s-%s-%s\n",os.c_str(),proc.c_str(),compiler.c_str());
    return EXIT_SUCCESS;
  }

  std::string expression = args[1];
  bool result;
  if(!evaluate_expression(os,proc,args[1],result)) {
    fprintf(stderr,"obuild_platform : problem in expression \"%s\".\n",
            expression.c_str());
    printf("no\n");
    return EXIT_FAILURE;
  } else {
    printf("%s\n",result?"yes":"no");
    return EXIT_SUCCESS;
  }
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
std::vector<std::string> Lib_smanip_words(
 const std::string& aString
,const std::string& aLimiter 
,bool aTakeEmpty // false
)
//////////////////////////////////////////////////////////////////////////////
//  If aLimiter is for exa "|" and for "xxx||xxx" :
//  - aTakeEmpty false : {"xxx","xxx"} will be created (and NOT {"xxx","","xxx"}).
//  - aTakeEmpty true : {"xxx","","xxx"} will be created.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::vector<std::string> words;
  if(aString=="") return words;
  std::string::size_type lim = (aTakeEmpty?0:1);
  if(aLimiter=="") {
    words.push_back(aString);
  } else {
    std::string::size_type l = aString.length();
    std::string::size_type llimiter = aLimiter.length();
    std::string::size_type pos = 0;
    while(1) {
      std::string::size_type index = aString.find(aLimiter,pos);
      if(index==std::string::npos){ // Last word.
        if((l-pos)>=lim) words.push_back(aString.substr(pos,l-pos));
        break;
      } else {
        //     abcxxxef
        //     0  3  67
        if((index-pos)>=lim) words.push_back(aString.substr(pos,index-pos));
        pos = index + llimiter;
      }
    }
  }
  return words;
}
