//  Mai/2007 : G.Barrand : creation.

//
// See doc/Game.txt for explanation
// of the controls and for a simple
// demo scenario.
//


#include <HEPVis/Gui/SoGui.h>

#include <Inventor/nodes/SoPointLight.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoDirectionalLight.h>
#include <Inventor/nodes/SoPerspectiveCamera.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoTranslation.h>
#include <Inventor/nodes/SoScale.h>
#include <Inventor/nodes/SoRotationXYZ.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/engines/SoTimeCounter.h>
#include <Inventor/engines/SoCalculator.h>

#include <HEPVis/SbMath.h>
#include <HEPVis/SbColors.h>
#include <HEPVis/SbTestPlotter.h>
#include <HEPVis/SbFile.h>
#include <HEPVis/SbString.h>
#include <HEPVis/SbArguments.h>
#include <HEPVis/misc/SoTools.h>
#include <HEPVis/nodes/SoGrid.h>
#include <HEPVis/nodes/SoAreaStyle.h>
#include <HEPVis/nodes/SoTextStyle.h>
#include <HEPVis/nodes/SoStyle.h>
#include <HEPVis/nodekits/SoDetectorExample.h>
#include <HEPVis/nodekits/SoDisplayRegion.h>
#include <HEPVis/nodekits/SoFormRegion.h>
#include <HEPVis/nodekits/SoSimpleButton.h>
#include <HEPVis/nodekits/SoPlotter.h>
#include <HEPVis/nodekits/SoDetectorTreeKit.h>

#include <HEPVis/SDL/viewers/SoSDL_Viewer.h>

#include <iostream>

extern "C" {
  void HEPVisUtils_initClasses();
  void HEPVisGeometry_initClasses();
  void HEPVisDetector_initClasses();
  //void HEPVisGUI_initClasses();
  void HEPVisPlotter_initClasses();
}

static void detPickCB(void*,void*);
static void activateCB(void*,SoNode*);
static void setCamera(SoPerspectiveCamera&,float);
static void unlockCB(void*);

#define ID_Quit 0
#define ID_RemoveManips 1
#define ID_Reset 2
#define ID_StarWars 3
#define ID_Sensitivity 4

//Darwin : "main" becomes "SDL_main" through a cpp macro in SDL.h.

class add_pickCB : public virtual SbVisitor {
public:
  virtual ~add_pickCB(){}
public:
  virtual SbBool visit(SoFullPath& aPath) {
    SoNode* node = aPath.getTail();
    if(!node) return TRUE;
    SoDetectorTreeKit* kit = (SoDetectorTreeKit*)node;
    kit->addPickCallback(fCB,fTag);
    return TRUE;
  }
public:
  SoCallbackListCB* fCB;
  void* fTag;
};

#ifdef HEPVisSDL
class GameViewer : public HEPVis::SoSDL_Viewer {
public:
  GameViewer(SoSDL_Widget* aWidget)
  : HEPVis::SoSDL_Viewer(aWidget)
  {}
  virtual ~GameViewer(){}
public:
  virtual SbBool processEvent(SDL_Event& aEvent,SbBool& aRender){
    if(processEventLock(aEvent,aRender)==TRUE) return TRUE;
    return HEPVis::SoSDL_Viewer::processEvent(aEvent,aRender);
  }
};
#else
#define GameViewer SoGuiExaminerViewer
#endif

static SoPerspectiveCamera* sCamera = 0;
static SoRotationXYZ* sRotation = 0;
static SoSeparator* sSceneGraph = 0;
static SoMaterial* sFloorMaterial = 0;
static GameViewer* sViewer = 0;
static SoSimpleButton* s_button_sensitivity = 0;

static int s_delta_step = 0;
static float s_delta_trans = 1.0F;
static float s_delta_rot = 1.0F; //degree

//////////////////////////////////////////////////////////////////////////////
int main( 
 int aArgc
,char* aArgv[]
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  SbArguments args(aArgc,aArgv);
  if(args.isAnArgument("-h")) {
    std::cout << "HEPVis::Game :" << std::endl;
    std::cout << "Usage :" << std::endl;
    std::cout << " OS> HEPVis_Game_<GUI> <options>" << std::endl;
    std::cout << "Options :" << std::endl;
    std::cout << " -h : to get help." << std::endl;
    std::cout << " -password=<string> : to lock screen." << std::endl;
    std::cout << " -not_full_screen : to not start in full screen." 
              << std::endl;
    std::cout << " -vide_width=<int> : video width (default is 640)." 
              << std::endl;
    std::cout << " -vide_height=<int> : video height (default is 480)." 
              << std::endl;
    return 0;
  }

  SbString password;
  if(args.find("-password",password)==FALSE) {
    SbString_getenv("HEPVIS_PASSWORD",password);
  }

  int video_width = 640;
 {SbString s;  
  if(args.find("-video_width",s)==TRUE) {
    int v;
    if(SbStringConvertToInt(s,v)==TRUE) video_width = v;
  }}
  int video_height = 480;
 {SbString s;  
  if(args.find("-video_height",s)==TRUE) {
    int v;
    if(SbStringConvertToInt(s,v)==TRUE) video_height = v;
  }}

  ///////////////////////////////////////////////
  // Init :
  ///////////////////////////////////////////////
  Uint32 flags = SDL_FULLSCREEN;
  //Uint32 flags = 0; //not full screen.
  if(args.isAnArgument("-not_full_screen")) flags = 0;

  SoGuiWidget topWidget = SoGui::init(video_width,video_height,flags);
  if(!topWidget) return 1;

  HEPVisUtils_initClasses();
  HEPVisGeometry_initClasses();
  HEPVisDetector_initClasses();
  //HEPVisGUI_initClasses();
  HEPVisPlotter_initClasses();

  SDL_Surface* screen = SoSDL::screen();
  if(!screen) return 1;
    
  sSceneGraph = new SoSeparator;
  sSceneGraph->ref();

  SoDisplayRegion* land = new SoDisplayRegion;
  //land->color.setValue(SbColor(0.8F,0.8F,0.8F));
  land->color.setValue(SbColor(0.5F,0.5F,0.5F));
  land->setCameraType(SoPerspectiveCamera::getClassTypeId());
  sSceneGraph->addChild(land);

  ///////////////////////////////////////////////
  // Camera (us) :
  ///////////////////////////////////////////////
  sCamera = (SoPerspectiveCamera*)land->getCamera();

  SoSeparator* scene = land->getStaticScene();

  // Light at ceil :
  SoPointLight* light = new SoPointLight;
  light->location.setValue(0,100,0); //y = 100.
  scene->addChild(light);

  ///////////////////////////////////////////////
  // Floor :
  ///////////////////////////////////////////////
  sFloorMaterial = new SoMaterial;
  if(password.getLength())
    sFloorMaterial->diffuseColor.setValue(1,0,0);
  else
    sFloorMaterial->diffuseColor.setValue(0.8F,0.8F,0.8F);
  scene->addChild(sFloorMaterial);

  SoGrid* floor = new SoGrid;
  float halfSize = 1000;
  floor->position.setValue(SbVec3f(-halfSize,0,halfSize)); //y = 0
  floor->X.setValue(SbVec3f(1,0,0));
  floor->Y.setValue(SbVec3f(0,0,-1));
  floor->width.setValue(2.0F*halfSize);
  floor->height.setValue(2.0F*halfSize);
  scene->addChild(floor);
  
  bool have_plotter = false;

  ///////////////////////////////////////////////
  // Detector :
  ///////////////////////////////////////////////
 {SoSeparator* soSep = new SoSeparator;
  scene->addChild(soSep);
  soSep->renderCaching.setValue(SoSeparator::ON);

  SoSeparator* separator = 0;

  SbString file;
  if(SbString_getenv("HEPVIS_FILE",file)==FALSE) {
    if(args.find("-file",file)==FALSE) {
      if(args.file(file)==FALSE) file = "out.iv";
    }
  }
  if(file.getLength()) {
    SbString name = SbFile::getFullName(file);
    SoInput soInput;
    if(soInput.openFile(name.getString())==TRUE) {
      separator = SoDB::readAll(&soInput);
    }
  }

  if(separator) {

    // LHCb setup :

    s_delta_trans = 0.6F;

    setCamera(*sCamera,20);

    SoTranslation* trans = new SoTranslation;
    trans->translation.setValue(SbVec3f(0,5,0)); //y = 5
    soSep->addChild(trans);

    SoScale* scale = new SoScale;
    float s = 0.001F;
    scale->scaleFactor.setValue(SbVec3f(s,s,s));
    soSep->addChild(scale);

    sRotation = new SoRotationXYZ; //For lock animation
    sRotation->axis.setValue(SoRotationXYZ::Y);
    //sRotation->angle.setValue(FM_PI_2);
    soSep->addChild(sRotation);

    soSep->addChild(separator);

    add_pickCB v;
    v.fCB = detPickCB;
    v.fTag = scene;
    SoTools::visit(*separator,SoDetectorTreeKit::getClassTypeId(),v);

  } else {

    have_plotter = true;

    s_delta_trans = 1.0F;

    setCamera(*sCamera,100);

    // Default scene :
    SoTranslation* trans = new SoTranslation;
    trans->translation.setValue(SbVec3f(0,20,0)); //y = 20
    soSep->addChild(trans);

    sRotation = new SoRotationXYZ; //For lock animation
    sRotation->axis.setValue(SoRotationXYZ::Y);
    //sRotation->angle.setValue(FM_PI_2);
    soSep->addChild(sRotation);

    SoDetectorExample* soDetExa = new SoDetectorExample;
    soSep->addChild(soDetExa);

    add_pickCB v;
    v.fCB = detPickCB;
    v.fTag = scene;
    SoTools::visit(*soDetExa,SoDetectorTreeKit::getClassTypeId(),v);

  }

  }

  ///////////////////////////////////////////////
  // Plotter :
  ///////////////////////////////////////////////
  if(have_plotter) 
 {SoSeparator* soSep = new SoSeparator;
  scene->addChild(soSep);

  SoTranslation* trans = new SoTranslation;
  trans->translation.setValue(SbVec3f(20,5+2,0));
  soSep->addChild(trans);

  SoScale* scale = new SoScale;
  scale->scaleFactor.setValue(SbVec3f(10,10,0.5));
  soSep->addChild(scale);

  SoPlotter* soPlotter = new SoPlotter;
  soPlotter->setName("plotter_3");
  SbTestHistogram1D* histo1D = new SbTestHistogram1D;
  soPlotter->setPlottableObject(new SbTestPlottableBins1D(*histo1D));
  soSep->addChild(soPlotter);

 {SoStyle* style = soPlotter->getGridStyle();
  style->visible.setValue(FALSE);}

 {SoStyle* style = soPlotter->getBinsStyle(0);
  //style->color.setValue(SbColor_navyblue);
  //style->modeling.setValue(SbModeling_boxes);
  style->modeling.setValue(SbModeling_top_lines);
  }

  } //End plotter.

  ///////////////////////////////////////////////
  // GUI in the scene :
  ///////////////////////////////////////////////
 {SoSeparator* soSep = new SoSeparator;
  scene->addChild(soSep);

/*
  SoTranslation* trans = new SoTranslation;
  trans->translation.setValue(SbVec3f(-1,2,z_cam*0.9));
  soSep->addChild(trans);

  SoScale* scale = new SoScale;
  scale->scaleFactor.setValue(SbVec3f(1,1,0.1));
  soSep->addChild(scale);

 {SoSimpleButton* button = new SoSimpleButton;
  button->getBackStyle()->color.setValue(SbColor_white);
  //button->getLabelStyle()->color.setValue(SbColor(0.7F,0,7F,0.7F));
  button->getLabelStyle()->fontName.setValue("Text3/defaultFont");
   //button->shadowThickness.setValue(0.1F);
   //button->marginWidth.setValue(0.1F);
   //button->marginHeight.setValue(0.1F);
  button->label.setValue("X");
  button->addCallback(activateCB,(void*)ID_RemoveManips);
  soSep->addChild(button);}

 {SoTranslation* trans = new SoTranslation;
  trans->translation.setValue(SbVec3f(-2,0,0));
  soSep->addChild(trans);}

 {SoSimpleButton* button = new SoSimpleButton;
  button->getBackStyle()->color.setValue(SbColor_white);
  //button->getLabelStyle()->color.setValue(SbColor(0.7F,0,7F,0.7F));
  button->getLabelStyle()->fontName.setValue("Text3/defaultFont");
   //button->shadowThickness.setValue(0.1F);
   //button->marginWidth.setValue(0.1F);
   //button->marginHeight.setValue(0.1F);
  button->label.setValue("SW");
  //button->addCallback(activateCB,(void*)ID_StarWars);
  soSep->addChild(button);}
*/

  } //End GUI in the scene.

  ///////////////////////////////////////////////
  // Shooting cross :
  ///////////////////////////////////////////////
  SoRegion* shootRegion = new SoRegion;
  float w = 0.008F;
  float h = 0.008F;
  shootRegion->origin.setValue(SbVec2f(0.5F*(1.0F-w),0.5F*(1.0F-h)));
  shootRegion->size.setValue(SbVec2f(w,h));
  shootRegion->visible.setValue(FALSE);
  sSceneGraph->addChild(shootRegion);

  ///////////////////////////////////////////////
  // GUI :
  ///////////////////////////////////////////////
  SoFormRegion* ui = new SoFormRegion;
  ui->origin.setValue(SbVec2f(0,0));
  //ui->size.setValue(SbVec2f(0.4F,0.1F));
  ui->size.setValue(SbVec2f(0.2F,0.05F));
  ui->columns.setValue(5);
  ui->rows.setValue(1);
  ui->visible.setValue(FALSE);
  sSceneGraph->addChild(ui);

 {SoSimpleButton* button = new SoSimpleButton;
  button->getBackStyle()->color.setValue(SbColor_white);
  //button->getLabelStyle()->color.setValue(SbColor(0.7F,0,7F,0.7F));
  button->getLabelStyle()->fontName.setValue("Text3/defaultFont");
   //button->shadowThickness.setValue(0.1F);
   //button->marginWidth.setValue(0.1F);
   //button->marginHeight.setValue(0.1F);
  button->label.setValue("Q");
  button->addCallback(activateCB,(void*)ID_Quit);
  ui->shapes.set1Value(ID_Quit,button);}

 {SoSimpleButton* button = new SoSimpleButton;
  button->getBackStyle()->color.setValue(SbColor_white);
  button->getLabelStyle()->fontName.setValue("Text3/defaultFont");
   //button->shadowThickness.setValue(0.1F);
   //button->marginWidth.setValue(0.1F);
   //button->marginHeight.setValue(0.1F);
  button->label.setValue("X");
  button->addCallback(activateCB,(void*)ID_RemoveManips);
  ui->shapes.set1Value(ID_RemoveManips,button);}

 {SoSimpleButton* button = new SoSimpleButton;
  button->getBackStyle()->color.setValue(SbColor_white);
  button->getLabelStyle()->fontName.setValue("Text3/defaultFont");
   //button->shadowThickness.setValue(0.1F);
   //button->marginWidth.setValue(0.1F);
   //button->marginHeight.setValue(0.1F);
  button->label.setValue("Reset");
  button->addCallback(activateCB,(void*)ID_Reset);
  ui->shapes.set1Value(ID_Reset,button);}

 {SoSimpleButton* button = new SoSimpleButton;
  button->getBackStyle()->color.setValue(SbColor_white);
  button->getLabelStyle()->fontName.setValue("Text3/defaultFont");
   //button->shadowThickness.setValue(0.1F);
   //button->marginWidth.setValue(0.1F);
   //button->marginHeight.setValue(0.1F);
  button->label.setValue("SW");
  button->addCallback(activateCB,(void*)ID_StarWars);
  ui->shapes.set1Value(ID_StarWars,button);}

 {s_button_sensitivity = new SoSimpleButton;
  s_button_sensitivity->getBackStyle()->color.setValue(SbColor_white);
  s_button_sensitivity->getLabelStyle()->fontName.setValue
    ("Text3/defaultFont");
   //button->shadowThickness.setValue(0.1F);
   //button->marginWidth.setValue(0.1F);
   //button->marginHeight.setValue(0.1F);
  s_button_sensitivity->label.setValue("0");
  s_button_sensitivity->addCallback(activateCB,(void*)ID_Sensitivity);
  ui->shapes.set1Value(ID_Sensitivity,s_button_sensitivity);}

  // Viewer :
  sViewer = new GameViewer(topWidget);
  sViewer->setSceneGraph(sSceneGraph);
  sViewer->setCamera(sCamera);
  sViewer->setHeadlight(FALSE);
  //sViewer->viewAll();
  sViewer->setShootRegion(shootRegion);
  sViewer->setOptionsRegion(ui);
  // adapt delta movement to the scene :
  sViewer->setDefaultDeltaTranslation(s_delta_trans);
  sViewer->setDefaultDeltaRotation(s_delta_rot * 3.1416F/180.0F);
  sViewer->checkFloor(TRUE);
  sViewer->setFloor(0.01F); //should take into account the camera height.
  sViewer->show();
    
  SoSDL::show(topWidget);

#ifdef HEPVisSDL
  if(password.getLength()) {
    sViewer->lock(password,unlockCB);

    SoTimeCounter* timeCounter = new SoTimeCounter;
    //sTimeCounter->on.setValue(FALSE);
    timeCounter->min.setValue(0); 
    timeCounter->max.setValue(200); 
    timeCounter->frequency.setValue(0.02F);
    SoCalculator* calculator = new SoCalculator();
    calculator->a.connectFrom(&(timeCounter->output));
    calculator->expression.set1Value(0,"oa=0.0314*a");
    sRotation->angle.connectFrom(&(calculator->oa));
  }
#endif

  SoSDL::mainLoop();
    
  sSceneGraph->unref();

  delete sViewer;
  sViewer = 0;

#ifdef HEPVisSDL
  SoGui::clear();  
#endif
	
  return 0;
}
//////////////////////////////////////////////////////////////////////////////
static void detPickCB(
 void* aTag
,void* aNode
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  SoDetectorTreeKit* soDet = (SoDetectorTreeKit*)((SoNode*)aNode);

/*
  //FIXME : highlight/unhighlight logic
  SoMaterial* material = 
    (SoMaterial*)sCurrentDet->getPart("appearance.material",TRUE);
  SbColor color = material->diffuseColor[0];
  if(color[0]==1) {
    material->diffuseColor.setValue(0,1,0);
  } else {
    material->diffuseColor.setValue(1,0,0);
  }
*/

  printf("HEPVis_Game::detPickCB : %s\n",soDet->getName().getString());

  SoNode* node = soDet->getPart("transform",TRUE);
  if(node) {
    SoNode* sceneGraph = (SoNode*)aTag;
    SoTools::setManip(sceneGraph,(SoTransform*)node);
    SDL_ShowCursor(1);
  }
}
static SoTimeCounter* sTimeCounter = 0;
//////////////////////////////////////////////////////////////////////////////
static void activateCB (
 void* aUserData
,SoNode*
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  unsigned long what = (unsigned long)aUserData;
  if(what==ID_Quit) {
    SoSDL::toQuit();
  } else if(what==ID_Reset) {

    sCamera->viewportMapping.setValue(SoCamera::ADJUST_CAMERA);
    float z_cam = 100;
    sCamera->position.setValue(0,2,z_cam); //y = 2 (meter)
    sCamera->focalDistance.setValue(z_cam);
    sCamera->orientation.setValue(SbRotation(SbVec3f(0,1,0),0));
    sCamera->nearDistance.setValue(z_cam*0.01F);
    sCamera->farDistance.setValue(100*z_cam);
    sCamera->heightAngle.setValue(FM_PI/4.0F);

  } else if(what==ID_RemoveManips) {

    SoTools::removeManips(*sSceneGraph);
    SDL_ShowCursor(0);

  } else if(what==ID_StarWars) {
    if(!sTimeCounter) {
#ifdef HEPVisSDL
      SbString file;
      if(SbString_getenv("HEPVIS_WAV",file)==TRUE) {
        SoGui::playAudioFile(file);  
      } else {
        SoGui::playAudioFile("starwars.wav");  
      }
#endif
      sTimeCounter = new SoTimeCounter;
      //sTimeCounter->on.setValue(FALSE);
      sTimeCounter->min.setValue(0); 
      sTimeCounter->max.setValue(1000); 
      sTimeCounter->frequency.setValue(0.02F);
      SoCalculator* calculator = new SoCalculator();
      calculator->a.connectFrom(&(sTimeCounter->output));
      calculator->expression.set1Value(0,"oA=vec3f(0,2,-100+a)");
      sCamera->position.connectFrom(&(calculator->oA));
      sCamera->orientation.setValue(SbRotation(SbVec3f(0,1,0),0));

    } else {
#ifdef HEPVisSDL
      SoSDL::clearAudio();
#endif
      sCamera->position.disconnect();
      sTimeCounter = 0;
    }

  } else if(what==ID_Sensitivity) {

    s_delta_step++;
    if(s_delta_step>10) s_delta_step = 0;

   {float fact = 1.0F/FPOW(10,s_delta_step);
    sViewer->setDefaultDeltaTranslation(fact * s_delta_trans);}

   {float fact = 1.0F/FPOW(2,s_delta_step);
    sViewer->setDefaultDeltaRotation(fact * s_delta_rot * 3.1416F/180.0F);}

   {SbString s;
    SbStringPrintF(s,32,"%d",s_delta_step);
    s_button_sensitivity->label.setValue(s);}

  //} else if(what==ID_AudioClear) {
  //  SoSDL::clearAudio();

  }
}

void setCamera(SoPerspectiveCamera& aCamera,float aZ) {
  aCamera.viewportMapping.setValue(SoCamera::ADJUST_CAMERA);
  aCamera.position.setValue(0,2,aZ); //y = 2 (meter)
  aCamera.focalDistance.setValue(aZ);
  aCamera.orientation.setValue(SbRotation(SbVec3f(0,1,0),0));
  aCamera.nearDistance.setValue(aZ*0.01F);
  aCamera.farDistance.setValue(100*aZ);
  aCamera.heightAngle.setValue(FM_PI/4.0F);
}

void unlockCB(void* aTag) {
  sFloorMaterial->diffuseColor.setValue(0.8F,0.8F,0.8F);
}
