/*
  Example of usage of SoPlotter.

  SoPlotter knows only the abstract classes :
   SbPlottableBins1D, SbPlottableCloud,...
 These classes permit SoPlotter to retreive
 informations about things to plot ; histogram,
 function,... The virtuality of the SbPlottable
 classes permits to plug many histogram packages
 to SoPlotter. 

  Given an histogram package <Xx> someone has to implement 
 the concrete class XxPlottableBins1D,... derived from
 SbPlottableBins1D...
 
  At run time someone create an histogram instance <Xx>Histogram
 of the package <Xx>, a corresponding instance <Xx>PlottableBins1D
 and give this instance to SoPlotter.

  SbTestPlotter.h contains a simple dummy histogram package.
*/
  

#include <HEPVis/Gui/SoGui.h>

// Inventor :
#include <Inventor/SbString.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoOrthographicCamera.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoDirectionalLight.h>

// HEPVisUtils :
#include <HEPVis/SbMath.h>
#include <HEPVis/SbStyles.h>
#include <HEPVis/SbColors.h>
#include <HEPVis/SbArguments.h>
#include <HEPVis/SbString.h>
#include <HEPVis/misc/SoTools.h>
#include <HEPVis/nodes/SoTextStyle.h>
#include <HEPVis/nodes/SoStyle.h>

// HEPVisPlotter :
#include <HEPVis/SbPlottableThings.h>
#include <HEPVis/nodekits/SoPlotter.h>
#include <HEPVis/nodekits/SoAxis.h>

extern "C" {
  void HEPVisUtils_initClasses();
  void HEPVisPlotter_initClasses();
}

#include <HEPVis/SbTestPlotter.h>

// Else :
#include <stdlib.h>
#include <iostream>

//////////////////////////////////////////////////////////////////////////////
int main (
 int aArgc
,char** aArgv
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  SbArguments args(aArgc,aArgv);

  if(args.isAnArgument("-h")) {
    std::cout << "HEPVis::Plotter :" << std::endl;
    std::cout << "Usage :" << std::endl;
    std::cout << " OS> Plotter<GUI> <options>" << std::endl;
    std::cout << "Options :" << std::endl;
    std::cout << " -h : to get help." << std::endl;
    std::cout << " -altrep : to produce an alternate representation." 
              << std::endl;
    std::cout << " -write : to write scene in a file (default is out.iv)." 
              << std::endl;
    std::cout << " -file : name of the file (with .iv or .wrl extension)." 
              << std::endl;
    std::cout << " -markers : to visualize bins with markers." << std::endl;
    std::cout << "For example :" << std::endl;
    std::cout << " OS> PlotterXt -altrep -write" << std::endl;
    std::cout << " UNIX> PlotterXt -altrep -write -file=out.wrl" << std::endl;
    std::cout << " UNIX> ivview out.iv" << std::endl;
    return 0;
  }

  // Do user data analysis (here very simple !) :
  SbTestHistogram1D* histo1D = new SbTestHistogram1D;

  SoGuiWidget widget = SoGui::init(aArgv[0]);
  if (widget==NULL) exit(1);

  HEPVisUtils_initClasses();
  HEPVisPlotter_initClasses();

  SoSeparator* sceneGraph = new SoSeparator;
  sceneGraph->ref();

  SoLightModel* lightModel = new SoLightModel;
  lightModel->model.setValue(SoLightModel::PHONG);
  //lightModel->model.setValue(SoLightModel::BASE_COLOR);
  sceneGraph->addChild(lightModel);

  SoDirectionalLight* light = new SoDirectionalLight;
  light->color.setValue(1,1,1);
  light->direction.setValue(SbVec3f(1,-1,-10));
  sceneGraph->addChild(light);

  SoOrthographicCamera* camera = new SoOrthographicCamera;
  //camera->viewportMapping.setValue(SoCamera::ADJUST_CAMERA);
  camera->viewportMapping.setValue(SoCamera::LEAVE_ALONE);
  camera->position.setValue(0,0,4);
  camera->orientation.setValue(SbRotation(SbVec3f(0,1,0),0));
  camera->height.setValue(1);
  camera->nearDistance.setValue(1);
  camera->farDistance.setValue(100);
  camera->focalDistance.setValue(4);
  sceneGraph->addChild(camera);

  SoPlotter* plotter = new SoPlotter;
  plotter->setName("plotter_3");
  plotter->setPlottableObject(new SbTestPlottableBins1D(*histo1D));

 {SoStyle* style = plotter->getWallStyle();
  style->modeling.setValue(SbModeling_boxes);
  style->color.setValue(SbColor_lightgrey);}

 {SoStyle* style = plotter->getBinsStyle(0);
  style->color.setValue(SbColor_navyblue);
  if(args.isAnArgument("-markers")) {
    style->modeling.setValue(SbModeling_points);
    style->markerStyle.setValue(SbMarkerPlus);
    style->markerSize.setValue(9);
  } else {
    //style->modeling.setValue(SbModeling_wire_boxes);
    style->modeling.setValue(SbModeling_boxes);
    style->hatching.setValue(1);
  }}

 {SoStyle* style = plotter->getRightHatchStyle(0);
  style->color.setValue(SbColor_white);}

  plotter->titleUp.setValue(FALSE);

 {SoTextStyle* style = plotter->getTitleStyle();
  style->color.setValue(SbColor_black);
  style->fontName.setValue("Hershey");
  //style->fontName.setValue("TTF/times"); //Does not have an alternateRep yet.
 }

 {SoTextStyle* style = plotter->getXAxis()->getLabelsStyle();
  style->color.setValue(SbColor_black);
  style->fontName.setValue("Hershey");
  //style->fontName.setValue("TTF/times"); //Does not have an alternateRep yet.
 }
 {SoTextStyle* style = plotter->getYAxis()->getLabelsStyle();
  style->color.setValue(SbColor_black);
  style->fontName.setValue("Hershey");
  //style->fontName.setValue("TTF/times"); //Does not have an alternateRep yet.
 }

  sceneGraph->addChild(plotter);

  if (args.isAnArgument("-write")) {
    SbBool altRep = args.isAnArgument("-altrep")?TRUE:FALSE;
    SbString file;
    args.find("-file",file);
    if(file=="") file = "out.iv";

    // force the construction of the plotter scene :
    plotter->raiseAllUpdateFlags();
    plotter->updateChildren();

    SoTools::write(*sceneGraph,file,"",FALSE,altRep);

    //SoTools::write(*(plotter->getTopSeparator()),file,"",FALSE,altRep);
  }

  ////////////////////////////////////////////////////////////

  SoGuiExaminerViewer* examinerViewer = new SoGuiExaminerViewer(widget);
  examinerViewer->setSize(SbVec2s(700,450));
  examinerViewer->setAutoClipping(FALSE);
  examinerViewer->setViewing(FALSE);
  examinerViewer->setHeadlight(FALSE);
  examinerViewer->setDecoration(FALSE);
  examinerViewer->setBackgroundColor(SbColor(1,1,1));
  examinerViewer->setSceneGraph(sceneGraph);
  examinerViewer->show();

  SoGui::show(widget);

  SoGui::mainLoop();

  //Cleanup
  sceneGraph->unref();
  delete examinerViewer;

  delete histo1D;

#ifdef HEPVisSDL
  SoGui::clear();  
#endif

  return EXIT_SUCCESS;
}
