// this :
#include <Lib/As.h>

#include <math.h> //pow

typedef std::complex<double> z;

//////////////////////////////////////////////////////////////////////////////
/// Some common matrices /////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////
//
//  En_k with k in [0,n(n-1)/2[
//
//  En_0 :        En_1 :         En_2 :            En_3
//   0  1  0...     0  0 -1...     0  0  0  0...     0  0  0  1... 
//  -1  0  0...     0  0  0...     0  0  1  0...     0  0  0  0...
//   0  0  0...     1  0  0...     0 -1  0  0...     0  0  0  0...
//   0  0  0...     0  0  0...     0  0  0  0...    -1  0  0  0...
//   ..........     ..........     .............     .............
//
//  with r<c r,c in [0,n[ :
//    k(r,c) = c*(c-1)/2 + r
//  Reverse, given k :
//    c is the first such that : c*(c-1)/2 <= k <= c*(c+1)/2 -1
//  then :
//    r = k - c*(c-1)/2
//
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
Lib::ADDd_E::ADDd_E(
 unsigned int aOrder
,unsigned int aIndex
) 
:ADD<double>(aOrder)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!initialize(aOrder,aIndex)) return; //FIXME : throw
}
//////////////////////////////////////////////////////////////////////////////
Lib::ADDd_E::~ADDd_E(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::ADDd_E::initialize(
 unsigned int aOrder
,unsigned int aIndex
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  unsigned int number = aOrder * (aOrder-1)/2;
  if(aIndex>number) return false;

  // Find column from index :
  bool found = false;
  unsigned int c = 0;
  unsigned int mn = 0;
  unsigned int mx = 0;
  for(c=1;c<aOrder;c++) {
    mn = c*(c-1)/2;
    mx = c*(c+1)/2-1;
    if((mn<=aIndex)&&(aIndex<=mx)) {
      found = true;
      break;
    }
  }
  if(!found) return false;
  // ok, then compute row :
  unsigned int r = aIndex - mn;
  //printf("debug : %d : r %d c %d : mn %d mx %d\n",aIndex,r,c,mn,mx);
  // then set value :
  double value = ::pow(-1.,(double)(r+c+1));
  // If one, then rot 3D vec csts are no more epsilon(i,j,k).
  //double value = 1;
  setValueNoCheck(r,c,value);
  setValueNoCheck(c,r,-value);

  unsigned int icheck = c*(c-1)/2+r;
  if(aIndex!=icheck) return false;
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::ADDd_E::groupConstants(
 unsigned int aOrder
,Lib::Array<double>& aConstants
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  unsigned int number = aOrder * (aOrder-1)/2;

  std::vector<unsigned int> orders(3,number);
  aConstants.configure(orders);

  std::vector< Lib::ADD<double> > Es;
  for(unsigned int index=0;index<number;index++)
    Es.push_back(Lib::ADDd_E(aOrder,index));

  Lib::ADD<double> zero(aOrder);

  for(unsigned int j=0;j<number;j++) {
    for(unsigned int k=j+1;k<number;k++) {
      Lib::ADD<double> cm = Lib::commutator(Es[j],Es[k]);

      // Find the result in the Es themselves :
      bool found = false;
      for(unsigned int l=0;l<number;l++) {
        if(cm.equal(zero)) {
          found = true; 
          break;
        } else if(cm.equal(Es[l])) {
          std::vector<unsigned int> is(3);
          is[0] = j;
          is[1] = k;
          is[2] = l;
          aConstants.setValue(is,1);
          is[0] = k;
          is[1] = j;
          is[2] = l;
          aConstants.setValue(is,-1);

          found = true; 
          break;
        } else {
          Lib::ADD<double> tmp(cm);
          tmp.multiply(-1.);
          if(tmp.equal(Es[l])) {
            std::vector<unsigned int> is(3);
            is[0] = j;
            is[1] = k;
            is[2] = l;
            aConstants.setValue(is,-1);
            is[0] = k;
            is[1] = j;
            is[2] = l;
            aConstants.setValue(is,1);

            found = true; 
            break;
          }
        }
      }
      if(!found) {
        aConstants.clear();
        return false;
      }
    }
  }
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::ADDd_E::metric(
 unsigned int aOrder
,Lib::ADD<double>& aMetric
) 
//////////////////////////////////////////////////////////////////////////////
// For order n : -2*I(n*(n-1)/2) for all orders.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  unsigned int number = aOrder * (aOrder-1)/2;

  std::vector< Lib::ADD<double> > Es;
  for(unsigned int index=0;index<number;index++)
    Es.push_back(Lib::ADDd_E(aOrder,index));

  aMetric.setOrder(number);

  for(unsigned int j=0;j<number;j++) {
    for(unsigned int k=0;k<number;k++) {
      Lib::ADD<double> mx = Es[j]*Es[k];
      if(!aMetric.setValue(j,k,mx.trace())) return false;
    }
  }

  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::ADDd_E::cartanMetric(
 unsigned int aOrder
,Lib::ADD<double>& aMetric
) 
//////////////////////////////////////////////////////////////////////////////
// For order n : -(2*n-4))*I(n*(n-1)/2)
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  unsigned int number = aOrder * (aOrder-1)/2;

  Lib::Array<double> csts;
  if(!Lib::ADDd_E::groupConstants(aOrder,csts)) {
    aMetric.clear();
    return false;
  }

  // Fill the adjoint representation :
  //      r        r
  //  (Xk)  = cst
  //      c      kc
  std::vector< Lib::ADD<double> > xs;
  for(unsigned int index=0;index<number;index++) {
    Lib::ADD<double> x(number);
    for(unsigned int j=0;j<number;j++) {
      for(unsigned int k=0;k<number;k++) {
        std::vector<unsigned int> is(3);
        is[0] = index;
        is[1] = k;
        is[2] = j;
        double value;
        if(!csts.value(is,value)) return false;
        if(!x.setValue(j,k,value)) return false;
      }
    }
    xs.push_back(x);
  }

  aMetric.setOrder(number);

  for(unsigned int j=0;j<number;j++) {
    for(unsigned int k=0;k<number;k++) {
      Lib::ADD<double> mx = xs[j]*xs[k];
      if(!aMetric.setValue(j,k,mx.trace())) return false;
    }
  }

  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::ADDd_E::casimir(
 unsigned int aOrder
,Lib::ADD<double>& aCasimir
) 
//////////////////////////////////////////////////////////////////////////////
//  For order n : (n-1)/2 * I(n)
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  unsigned int number = aOrder * (aOrder-1)/2;

  Lib::ADD<double> metric(number);
  if(!Lib::ADDd_E::metric(aOrder,metric)) {
    aCasimir.clear();
    return false;
  }

  Lib::ADD<double> metric_inv(number);
  if(!metric.invert(metric_inv)) return false;

  aCasimir.setOrder(aOrder);

  std::vector< Lib::ADD<double> > Es;
  for(unsigned int index=0;index<number;index++)
    Es.push_back(Lib::ADDd_E(aOrder,index));

  for(unsigned int j=0;j<number;j++) {
    for(unsigned int k=0;k<number;k++) {
      Lib::ADD<double> mx = Es[j]*Es[k];
      mx.multiply(metric_inv.value(j,k));
      aCasimir.add(mx);
    }
  }

  return true;
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
Lib::A666d_LorentzGroupConstants::A666d_LorentzGroupConstants(
) 
//////////////////////////////////////////////////////////////////////////////
// From the vector representation of the Lorentz group :
//   VecRep_k = eta * E4_k  
// with k in [0,6[
// We must have :
//  [VecRep_j,VecRep_k] = Lorentz_Group_const(j,k,l) * VecRep_l
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  setValue(0,1,2,1);
  setValue(0,2,1,1);
  setValue(0,3,4,1);
  setValue(0,4,3,1);

  setValue(1,2,0,-1);
  setValue(1,3,5,1);
  setValue(1,5,3,1);

  setValue(2,4,5,-1);
  setValue(2,5,4,1);

  setValue(3,4,0,-1);
  setValue(3,5,1,-1);

  setValue(4,5,2,-1);

  // Antysym parts :
  setValue(1,0,2,-1);
  setValue(2,0,1,-1);
  setValue(3,0,4,-1);
  setValue(4,0,3,-1);

  setValue(2,1,0,1);
  setValue(3,1,5,-1);
  setValue(5,1,3,-1);

  setValue(4,2,5,1);
  setValue(5,2,4,-1);

  setValue(4,3,0,1);
  setValue(5,3,1,1);

  setValue(5,4,2,1);
}
//////////////////////////////////////////////////////////////////////////////
Lib::A66d_LorentzCartanMetric::A66d_LorentzCartanMetric(
) 
:ADD<double>(6)
//////////////////////////////////////////////////////////////////////////////
// From the vector representation of the Lorentz group :
//   VecRep_k = eta * E4_k  
// with k in [0,6[
//
// Should be :
//   4  0  0  0  0  0 
//   0  4  0  0  0  0
//   0  0 -4  0  0  0
//   0  0  0  4  0  0
//   0  0  0  0 -4  0
//   0  0  0  0  0 -4
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::A666d_LorentzGroupConstants csts;

  unsigned int number = 6;

  // Fill the adjoint representation :
  //      r        r
  //  (Xk)  = cst
  //      c      kc
  std::vector< Lib::ADD<double> > xs;
  for(unsigned int index=0;index<number;index++) {
    Lib::ADD<double> x(number);
    for(unsigned int j=0;j<number;j++) {
      for(unsigned int k=0;k<number;k++) {
        if(!x.setValue(j,k,csts.value(index,k,j))) return;
      }
    }
    xs.push_back(x);
  }

  for(unsigned int j=0;j<number;j++) {
    for(unsigned int k=0;k<number;k++) {
      Lib::ADD<double> mx = xs[j]*xs[k];
      setValue(j,k,mx.trace());
    }
  }

}
//////////////////////////////////////////////////////////////////////////////
Lib::A66d_LorentzVecMetric::A66d_LorentzVecMetric(
) 
:ADD<double>(6)
//////////////////////////////////////////////////////////////////////////////
// From the vector representation of the Lorentz group :
//   VecRep_k = eta * E4_k  
// with k in [0,6[
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  setValue(0,0, 2);
  setValue(1,1, 2);
  setValue(2,2,-2);
  setValue(3,3, 2);
  setValue(4,4,-2);
  setValue(5,5,-2);
}
//////////////////////////////////////////////////////////////////////////////
Lib::A66z_LorentzDiracMetric::A66z_LorentzDiracMetric(
) 
:ADD<z>(6)
//////////////////////////////////////////////////////////////////////////////
// From the dirac spinor representation of the Lorentz group :
//   SpinRep_k = 1/8 * [gamma_mu,gamma_nu] * E4_k_mu_nu
// with k in [0,6[
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  setValue(0,0, 1);
  setValue(1,1, 1);
  setValue(2,2,-1);
  setValue(3,3, 1);
  setValue(4,4,-1);
  setValue(5,5,-1);
}
//////////////////////////////////////////////////////////////////////////////
Lib::A66z_LorentzWeylMetric::A66z_LorentzWeylMetric(
) 
:ADD<z>(6)
//////////////////////////////////////////////////////////////////////////////
// From the weyl spinor representation of the Lorentz group :
//   SpinRep_k = (pauli_1/2,pauli_2/2,pauli_3/2,
//                i*pauli_1/2,i*pauli_2/2,i*pauli_3/2)
//   k in [0,6[
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  setValue(0,0, 0.5);
  setValue(1,1, 0.5);
  setValue(2,2,-0.5);
  setValue(3,3, 0.5);
  setValue(4,4,-0.5);
  setValue(5,5,-0.5);

  z zi(0,1);
  setValue(0,5, zi*0.5);
  setValue(1,4,-zi*0.5);
  setValue(2,3, zi*0.5);
  setValue(3,2, zi*0.5);
  setValue(4,1,-zi*0.5);
  setValue(5,0, zi*0.5);

}
//////////////////////////////////////////////////////////////////////////////
Lib::ADDd_PoincareCartanMetric::ADDd_PoincareCartanMetric(
 unsigned int aOrder
) 
:ADD<double>(aOrder*(aOrder+1)/2)
//////////////////////////////////////////////////////////////////////////////
// Poincare group in "aOrder" dimension with delta metric (identity).
//
// From the operator representation :
//   OpRep_k = 1/2 * J_mu_nu * En_k_mu_nu   with k in [0,n*(n-1)/2[
//   OpRep_alpha = D_alpha                  with alpha in [0,n[
// With :
//   J_mu_nu =   delta_mu_alpha * coord_alpha o D_nu 
//             - delta_nu_alpha * coord_alpha o D_mu
// Not null group constants are :
//   c(j,k,l) = e_n(j,k,l)
//   c(j,alpha,beta) = En_j_beta_alpha
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  unsigned int number = aOrder*(aOrder+1)/2;

  // Fill the adjoint representation :
  //      c        c
  //  (Xa)  = cst
  //      b      ab
  // a,b,c in [0,n*(n+1)/2[

  std::vector< Lib::ADD<double> > xs;

 {Lib::Array<double> csts;
  if(!Lib::ADDd_E::groupConstants(aOrder,csts)) return;
  unsigned int n = aOrder*(aOrder-1)/2;
  for(unsigned int index=0;index<n;index++) {
    Lib::ADD<double> x(number);

   {for(unsigned int j=0;j<n;j++) {
      for(unsigned int k=0;k<n;k++) {
        std::vector<unsigned int> is(3);
        is[0] = index;
        is[1] = k;
        is[2] = j;
        double value;
        if(!csts.value(is,value)) return;
        if(!x.setValue(j,k,value)) return;
      }
    }}

   {Lib::ADDd_E E(aOrder,index);
    for(unsigned int j=0;j<aOrder;j++) {
      for(unsigned int k=0;k<aOrder;k++) {

        double value = E.value(j,k);

        if(!x.setValue(n+j,n+k,value)) return;
      }
    }}

    xs.push_back(x);
  }}

 {for(unsigned int index=0;index<aOrder;index++) {
    Lib::ADD<double> x(number);
    xs.push_back(x);
  }}

  // Set cartan killing metric :
 {for(unsigned int j=0;j<number;j++) {
    for(unsigned int k=0;k<number;k++) {
      Lib::ADD<double> mx = xs[j]*xs[k];
      if(!setValue(j,k,mx.trace())) return;
    }
  }}

}

/// To test compilation.

void Lib_As_test() {
  Lib::ADDd_E a1(3,0);
}


