//
//  All functions here should be OnX callbacks, that is to say
// functions with signature :
//   extern "C" {
//     void callback_without_arguments(Slash::UI::IUI&);
//     void callback_with_arguments(Slash::UI::IUI&,
//                                  const std::vector<std::string>&);
//   }
//

//
// HEPVis and Inventor related callbacks.
//

// Helpers :
#include <OnX/Helpers/OnX.h>
#include <OnX/Helpers/Inventor.h>

// Slash :
#include <Slash/Core/IWriter.h>
#include <Slash/UI/ICallbackData.h>

// Lib : 
#include <Lib/System.h>
#include <Lib/smanip.h>
#include <Lib/sout.h>
#include <Lib/Out.h>
#include <Lib/Arguments.h>

// OnX :
#include <OnX/Interfaces/IIB.h>

// Inventor :
#include <Inventor/errors/SoDebugError.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoScale.h>
#include <Inventor/nodes/SoOrthographicCamera.h>
#include <Inventor/nodes/SoPerspectiveCamera.h>
#include <Inventor/actions/SoCallbackAction.h>
#include <Inventor/actions/SoSearchAction.h>

// HEPVis :
#include <HEPVis/misc/SoTools.h>
#include <HEPVis/nodes/SoHighlightMaterial.h>
#include <HEPVis/nodes/SoCoordinateAxis.h>
#include <HEPVis/nodes/SoStyle.h>
#include <HEPVis/nodes/SoImage.h>
#include <HEPVis/nodes/SoTextStyle.h>
#include <HEPVis/nodes/SoViewportRegion.h>
#include <HEPVis/nodekits/SoPage.h>
#include <HEPVis/nodekits/SoPlotter.h>
#include <HEPVis/nodekits/SoRegion.h>
#include <HEPVis/nodekits/SoPlotterRegion.h>
#include <HEPVis/nodekits/SoDisplayRegion.h>
#include <HEPVis/nodekits/SoImageRegion.h>
#include <HEPVis/nodekits/SoTextRegion.h>
#include <HEPVis/nodekits/SoDetectorExample.h>
#include <HEPVis/nodekits/SoRulerRegion.h>
#include <HEPVis/actions/SoCounterAction.h>

inline IIB& toIB(Slash::UI::IUI& aUI) {
  return *Lib_Cast(aUI,IIB);
}

struct Counter {
 int fTriangles;
 int fLineSegments;
 int fPoints;
};

static void CountTrianglesCB(
 void* userData
,SoCallbackAction*
,const SoPrimitiveVertex*
,const SoPrimitiveVertex*,
const SoPrimitiveVertex*)
{
  Counter* counter = (Counter*)userData;
  counter->fTriangles++;
}

static void CountLineSegmentsCB(
 void* userData
,SoCallbackAction*
,const SoPrimitiveVertex*
,const SoPrimitiveVertex*)
{
  Counter* counter = (Counter*)userData;
  counter->fLineSegments++;
}

static void CountPointsCB(
 void* userData
,SoCallbackAction*
,const SoPrimitiveVertex*)
{
  Counter* counter = (Counter*)userData;
  counter->fPoints++;
}

//////////////////////////////////////////////////////////////////////////////
inline void page_dressCurrentRegionWithRulers(
 SoPage& aPage
,double aRulerSize = 0.9
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  SoRegion* soRegion = aPage.currentRegion();
  if(!soRegion) return;
  if(!soRegion->isOfType(SoDisplayRegion::getClassTypeId())) {
    SoDebugError::postInfo("page_dressCurrentRegionWithRulers",
      "current region is not a SoDisplayRegion.");
    return;
  }

  float w = float(aRulerSize);
  float h = float(aRulerSize);

  // Visualization region :
  SoViewportRegion* vpr = soRegion->getViewportRegion();
  vpr->setPositionPercent(1-w,1-h);
  vpr->setSizePercent(w,h);

  // Horizontal ruler region :
  SoRulerRegion* hRuler = 
    (SoRulerRegion*)aPage.createRegion("SoRulerRegion",1-w,0,w,1-h);
  if(!hRuler) return;
  hRuler->getViewportRegion()->backgroundColor.setValue(SbColor(1,1,1));

  // Connect the display region to the ruler :
  hRuler->setMeasuredRegion(soRegion);

  // Vertical ruler region :
  SoRulerRegion* vRuler = 
    (SoRulerRegion*)aPage.createRegion("SoRulerRegion",0,1-h,1-w,h);
  if(!vRuler) return;
  vRuler->getViewportRegion()->backgroundColor.setValue(SbColor(1,1,1));

  // Connect the display region to the ruler :
  vRuler->setMeasuredRegion(soRegion);

  // Set display region as current :
  aPage.highlightRegion(soRegion);

}
//////////////////////////////////////////////////////////////////////////////
inline void page_removeRulers(
 SoPage& aPage
) 
//////////////////////////////////////////////////////////////////////////////
// Find rulers and remove them.
// Map the first SoDisplayRegion to the page.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  SoRegion* soRegion = aPage.currentRegion();
  if(!soRegion) return;
  // Map first display region to page.
  std::vector<SoRegion*> torm;
 {int number = aPage.getNumberOfRootRegions();
  for(int index=0;index<number;index++) {
    SoRegion* soRegion = aPage.getRootRegion(index);
    if(soRegion->isOfType(SoDisplayRegion::getClassTypeId())) {
      SoViewportRegion* vpr = soRegion->getViewportRegion();
      vpr->setPositionPercent(0,0);
      vpr->setSizePercent(1,1);
    } else if(soRegion->isOfType(SoRulerRegion::getClassTypeId())) {
      torm.push_back(soRegion);
    }
  }}
 {unsigned int number = torm.size();
  for(unsigned int index=0;index<number;index++) {
    aPage.deleteRegion(torm[index]);
  }}
}

extern "C" {

//////////////////////////////////////////////////////////////////////////////
void OnX_Page_Clone_Viewer_1(
 Slash::UI::IUI& aUI
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  int n;
  if(!ui_parameterValue(aUI,"ViewerTabStack.itemCount",n)) return;

  std::string name;
  if(!Lib::smanip::printf(name,64,"Viewer_%d",n+1)) return;

  //printf("debug : clone Viewer_1\n");
  aUI.setParameter("Viewer_1.tabLabel",name);
  aUI.copyComponent(name,"Viewer_1","ViewerTabStack");
  aUI.setParameter("Viewer_1.tabLabel","Viewer_1");

  Slash::UI::IWidget* widget = aUI.findWidget(name);
  if(!widget) return;
  Slash::UI::ISoViewer* soViewer = Lib_Cast(*widget,Slash::UI::ISoViewer);
  if(!soViewer) return;
  aUI.setCurrentWidget(widget);
}
//////////////////////////////////////////////////////////////////////////////
void OnX_Page_New_Page(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string sclass = "PageViewer";
  if(aArgs.size()==1) sclass = aArgs[0];

  int n;
  if(!ui_parameterValue(aUI,"ViewerTabStack.itemCount",n)) return;

  std::string name;
  if(!Lib::smanip::printf(name,64,"Viewer_%d",n+1)) return;

  //printf("debug : create a new PageViewer\n");
  aUI.createComponent(name,sclass,"ViewerTabStack");

  aUI.setCallback(name,"destroy","DLD","OnX page_destroy");
  aUI.setCallback(name,"collect","DLD","OnX viewer_collect");
  aUI.setCallback(name,"popup","DLD","OnX viewer_popup");

  aUI.setParameter(name+".popupItems","No highlighted\nDump highlighted");

  Slash::UI::IWidget* widget = aUI.findWidget(name);
  if(!widget) return;
  Slash::UI::ISoViewer* soViewer = Lib_Cast(*widget,Slash::UI::ISoViewer);
  if(!soViewer) return;
  aUI.setCurrentWidget(widget);

  // Add in the Viewers pdm :
  if(aUI.isComponent("viewers")) {
    std::string s = "viewers_item_" + name;
    aUI.createComponent(s,"MenuItem","viewers");
    aUI.setParameter(s+".label",name);
    aUI.setCallback(s,"activate","DLD","OnX viewer_set_current "+name);
  }
  
}
//////////////////////////////////////////////////////////////////////////////
void OnX_Page_Delete_Page(
 Slash::UI::IUI& aUI
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Slash::UI::IWidget* widget = aUI.currentWidget();
  if(!widget) return;
  //FIXME : have a aUI.destroyWidget(widget);
  aUI.destroyComponent(widget->name());
}
//////////////////////////////////////////////////////////////////////////////
/// ui////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
void ui_createComponent(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// arguments : 
//   -class : if not, create a PageViewer.
//   -name : if not, generate one.
//   -parent : if not, take "ViewerTabStack".
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Arguments args(aArgs);
  std::string sclass;
  if(!args.find("-class",sclass)) sclass = "PageViewer";
  std::string sparent;
  if(!args.find("-parent",sparent)) sparent = "ViewerTabStack";
  std::string sname;
  if(!args.find("-name",sname)) {
    if(sparent=="ViewerTabStack") {
      int n;
      if(!ui_parameterValue(aUI,"ViewerTabStack.itemCount",n)) {
        Lib::Out out(aUI.session().printer());
        out << "ui_createComponent :"
            << " can't get ViewerTabStack.itemCount value."
            << Lib::endl;
        return;
      }
      if(!Lib::smanip::printf(sname,64,"Viewer_%d",n+1)) {
        Lib::Out out(aUI.session().printer());
        out << "ui_createComponent :"
            << " problem with Lib::printf."
            << Lib::endl;
        return;
      }
    } else {
      sname = toIB(aUI).createWidgetName();
    }
  }

  if(!aUI.createComponent(sname,sclass,sparent)) {
    Lib::Out out(aUI.session().printer());
    out << "ui_createComponent :"
        << " problem when creating component."
        << Lib::endl;
    return;
  }

  Slash::UI::IWidget* widget = aUI.findWidget(sname);
  if(widget) {
    // If a PageViewer or an OpenGLArea, set it as current widget.
    Slash::UI::ISoViewer* soViewer = Lib_Cast(*widget,Slash::UI::ISoViewer);
    if(soViewer) {
      aUI.setCurrentWidget(widget);
    } else if(sclass=="OpenGLArea") {
      aUI.setCurrentWidget(widget);
    }
  }

  /*
  Slash::UI::IWidgetClass* wclass = aUI.findWidgetClass(sclass);
  if(!wclass) {
    Lib::Out out(aUI.session().printer());
    out << "ui_createComponent :"
        << " widget class " << Lib::sout(sclass) << " not found." 
        << Lib::endl;
    return;
  }

  Slash::UI::IWidget* wparent = aUI.findWidget(sparent);
  if(!wparent) {
    Lib::Out out(aUI.session().printer());
    out << "ui_createComponent :"
        << " parent widget " << Lib::sout(sparent) << " not found." 
        << Lib::endl;
    return;
  }

  Slash::UI::IWidget* widget = aUI.createWidget(*wclass,sname,*wparent);
  if(!widget) {
    Lib::Out out(aUI.session().printer());
    out << "ui_createComponent :"
        << " can't create widget " << Lib::sout(sname)
        << Lib::dot_endl;
    return;
  }

  // If a PageViewer, set it as current viewer.
  Slash::UI::ISoViewer* soViewer = Lib_Cast(*widget,Slash::UI::ISoViewer);
  if(soViewer) aUI.setCurrentWidget(soViewer);*/

}
//////////////////////////////////////////////////////////////////////////////
/// page /////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
void page_next(
 Slash::UI::IUI& aUI
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;
  SoRegion* highlighted = soPage->getHighlightedRegion();
  if(!highlighted) return;
  SoRegion* nextSoRegion = soPage->getNextRootRegion(highlighted);
  if(!nextSoRegion) return;
  soPage->highlightRegion(nextSoRegion);
}
//////////////////////////////////////////////////////////////////////////////
void page_previous(
 Slash::UI::IUI& aUI
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;
  SoRegion* highlighted = soPage->getHighlightedRegion();
  if(!highlighted) return;
  SoRegion* prevSoRegion = soPage->getPreviousRootRegion(highlighted);
  if(!prevSoRegion) return;
  soPage->highlightRegion(prevSoRegion);
}
//////////////////////////////////////////////////////////////////////////////
void page_clear(
 Slash::UI::IUI& aUI
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;
  soPage->clearRegions();
}
//////////////////////////////////////////////////////////////////////////////
void page_refresh(
 Slash::UI::IUI& aUI
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;
  soPage->refreshRegions();
}
//////////////////////////////////////////////////////////////////////////////
void page_deleteRegions(
 Slash::UI::IUI& aUI
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;
  soPage->deleteRegions();
}
//////////////////////////////////////////////////////////////////////////////
void page_deleteRegion(
 Slash::UI::IUI& aUI
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;
  SoRegion* highlighted = soPage->getHighlightedRegion();
  if(!highlighted) return;
  if(soPage->isRootRegion(highlighted)==FALSE) return;
  soPage->deleteHighlightedRegion();
}
//////////////////////////////////////////////////////////////////////////////
void page_currentRegionToPage(
 Slash::UI::IUI& aUI
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;
  SoRegion* highlighted = soPage->getHighlightedRegion();
  if(!highlighted) return;
  if(soPage->isRootRegion(highlighted)==FALSE) return;
  soPage->mapHighlightedRegionToPage();
}
//////////////////////////////////////////////////////////////////////////////
void page_currentRegionOnTop(
 Slash::UI::IUI& aUI
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;
  SoRegion* highlighted = soPage->getHighlightedRegion();
  if(!highlighted) return;
  if(soPage->isRootRegion(highlighted)==FALSE) return;
  soPage->setRegionOnTop(highlighted);
}
//////////////////////////////////////////////////////////////////////////////
void page_addRegions(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aArgs.size()!=2) return;

  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;

  int cols;
  if(!Lib::smanip::toint(aArgs[0],cols)) return;
  int rows;
  if(!Lib::smanip::toint(aArgs[1],rows)) return;

  soPage->addRegions("SoDisplayRegion",cols,rows);
}
//////////////////////////////////////////////////////////////////////////////
void page_createRegion(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aArgs.size()!=5) return;

  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;

  std::string sclass = aArgs[0];

  double x;
  if(!Lib::smanip::todouble(aArgs[1],x)) return;
  double y;
  if(!Lib::smanip::todouble(aArgs[2],y)) return;
  double w;
  if(!Lib::smanip::todouble(aArgs[3],w)) return;
  double h;
  if(!Lib::smanip::todouble(aArgs[4],h)) return;

  SoRegion* soRegion = 
    soPage->createRegion(sclass.c_str(),(float)x,(float)y,(float)w,(float)h);
  if(!soRegion) return;
  soPage->setCurrentRegion(soRegion);
}
//////////////////////////////////////////////////////////////////////////////
void page_createRegions(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aArgs.size()!=4) return;

  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;

  int cols;
  if(!Lib::smanip::toint(aArgs[0],cols)) return;
  int rows;
  if(!Lib::smanip::toint(aArgs[1],rows)) return;
  int index;
  if(!Lib::smanip::toint(aArgs[2],index)) return;

  std::string sclass = aArgs[3];

  soPage->createRegions(sclass.c_str(),cols,rows,index);
}
//////////////////////////////////////////////////////////////////////////////
void page_setParameter(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// args[0] what
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;
  if(!aArgs.size()) return;
  if(aArgs.size()==1) {
    SoTools::setKitPartValue(soPage,aArgs[0].c_str(),"");
  } else {
    std::string value = ui_cbk(aUI,aArgs[1]);
    for(unsigned int index=2;index<aArgs.size();index++) 
      value += " " + ui_cbk(aUI,aArgs[index]);
    SoTools::setKitPartValue(soPage,aArgs[0].c_str(),value.c_str());
  }
}
//////////////////////////////////////////////////////////////////////////////
void page_toggleParameter(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// args[0] what
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!aArgs.size()) return;
  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;
  SoTools::toggleKitPartValue(soPage,aArgs[0].c_str());
}
//////////////////////////////////////////////////////////////////////////////
void page_setRootRegionsHighlightable(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// Highlight mechanism on regions may be disconnected (for exa in OpenPAW).
// The below restore it.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!aArgs.size()) return;
  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;
  bool b;
  if(!Lib::smanip::tobool(aArgs[0],b)) return;
  soPage->setRootRegionsHighlightable(b?TRUE:FALSE);
}
//////////////////////////////////////////////////////////////////////////////
void page_dressCurrentRegionWithRulers(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// aArgs[0] ruler size (default is 0.9)
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  double rulerSize = 0.9;
  if(aArgs.size()) {
    if(!Lib::smanip::todouble(aArgs[0],rulerSize)) return;
  }
  Slash::UI::ISoViewer* soViewer = ui_SoViewer(aUI);
  if(!soViewer) return;
  soViewer->setBackgroundColor(1,1,1);
  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;
  page_dressCurrentRegionWithRulers(*soPage,rulerSize);
}
//////////////////////////////////////////////////////////////////////////////
void page_removeRulers(
 Slash::UI::IUI& aUI
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;
  page_removeRulers(*soPage);
}
//////////////////////////////////////////////////////////////////////////////
/// region ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
void region_setNode(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// aArgs[0] what [part,name,type]
// aArgs[1] searched
// args[2,n-1] : partial value
//  value is built with : args[2] + " " + args[2] + ..... + args[n-1]
// Example : 
//   OnX region_setNode type SoCamera height 500 position 0 0 500
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;
  SoRegion* soRegion = soPage->currentRegion();
  if(!soRegion) return;
  SoNode* node = soRegion->findNode(aArgs[0].c_str(),aArgs[1].c_str());
  if(!node) {
    aUI.echo("node not found.");
    return;
  }
  std::string value = ui_cbk(aUI,aArgs[2]);
  for(unsigned int index=3;index<aArgs.size();index++) 
    value += " " + ui_cbk(aUI,aArgs[index]);
  node->set(value.c_str());
}
//////////////////////////////////////////////////////////////////////////////
void region_dumpCamera(
 Slash::UI::IUI& aUI
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;
  SoRegion* soRegion = soPage->currentRegion();
  if(!soRegion) return;
  SoCamera* camera = soRegion->getCamera();
  if(!camera) {
    aUI.echo("SoCamera not found.");
    return;
  }
  /*FIXME : not reliable. Sometime it works, sometime not !
  SbString ss;
  camera->get(ss);
  aUI.echo(ss.getString());*/

  std::string vpm;
  switch(camera->viewportMapping.getValue()){
    case SoCamera::CROP_VIEWPORT_FILL_FRAME:
      vpm = "CROP_VIEWPORT_FILL_FRAME";
      break;
    case SoCamera::CROP_VIEWPORT_LINE_FRAME:
      vpm = "CROP_VIEWPORT_LINE_FRAME";
      break;
    case SoCamera::CROP_VIEWPORT_NO_FRAME:
      vpm = "CROP_VIEWPORT_NO_FRAME";
      break;
    case SoCamera::ADJUST_CAMERA: 
      vpm = "ADJUST_CAMERA";
      break;
    case SoCamera::LEAVE_ALONE:
      vpm = "LEAVE_ALONE";
      break;
  }
  const SbVec3f&  position = camera->position.getValue();
  SbVec3f axis;
  float angle;
  camera->orientation.getValue().getValue(axis,angle);
  std::string sb;
  Lib::smanip::printf(sb,256,"\
viewportMapping %s\n\
position %g %g %g\n\
orientation %g %g %g %g\n\
aspectRatio %g\n\
nearDistance %g\n\
farDistance %g\n\
focalDistance %g\n"
,vpm.c_str()
,position[0],position[1],position[2]
,axis[0],axis[1],axis[2],angle
,camera->aspectRatio.getValue()
,camera->nearDistance.getValue()
,camera->farDistance.getValue()
,camera->focalDistance.getValue());
  std::string sh;
  if(camera->isOfType(SoOrthographicCamera::getClassTypeId())==TRUE) {
    Lib::smanip::printf(sh,64,"height %g",
      ((SoOrthographicCamera*)camera)->height.getValue());
  } else if(camera->isOfType(SoPerspectiveCamera::getClassTypeId())==TRUE) {
    Lib::smanip::printf(sh,64,"heightAngle %g",
      ((SoPerspectiveCamera*)camera)->heightAngle.getValue());
  }
  std::string sc = sb + sh;  
  aUI.echo(sc.c_str());
}
//////////////////////////////////////////////////////////////////////////////
void region_resetCamera(
 Slash::UI::IUI& aUI
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;
  SoRegion* soRegion = soPage->currentRegion();
  if(!soRegion) return;
  soRegion->resetCamera();
}
//////////////////////////////////////////////////////////////////////////////
void region_erase(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!aArgs.size()) return;
  if( (aArgs[0]=="static") || (aArgs[0]=="staticScene") ){
    SoPage* soPage = ui_SoPage(aUI);
    if(!soPage) return;
    SoRegion* soRegion = soPage->currentRegion();
    if(!soRegion) return;
    if(!soRegion->isOfType(SoDisplayRegion::getClassTypeId())) return;
    SoDisplayRegion* displayRegion = (SoDisplayRegion*)soRegion;
    SoSeparator* sep = (SoSeparator*)displayRegion->getStaticScene();
    if(sep) sep->removeAllChildren();
  } else if( (aArgs[0]=="dynamic") || (aArgs[0]=="dynamicScene") ) {
    SoPage* soPage = ui_SoPage(aUI);
    if(!soPage) return;
    SoRegion* soRegion = soPage->currentRegion();
    if(!soRegion) return;
    if(!soRegion->isOfType(SoDisplayRegion::getClassTypeId())) return;
    SoDisplayRegion* displayRegion = (SoDisplayRegion*)soRegion;
    SoSeparator* sep = (SoSeparator*)displayRegion->getDynamicScene();
    if(sep) sep->removeAllChildren();
  } else if(aArgs[0]=="highlighted") {
    data_collect(aUI.session(),"SceneGraph","highlight==true");
    data_destroy(aUI.session());
  } else if(aArgs[0]=="not_highlighted") {
    data_collect(aUI.session(),"SceneGraph","highlight==false");
    data_destroy(aUI.session());
  }
}
//////////////////////////////////////////////////////////////////////////////
void region_readScene(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!aArgs.size()) return;
  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;
  SoRegion* soRegion = soPage->currentRegion();
  if(!soRegion) return;
  std::string file = aArgs[0];
  std::string format = Lib::smanip::suffix(file);
  if(soRegion->readScene(file.c_str(),format.c_str(),"staticScene")==FALSE) {
    aUI.echo("Can't read file "+Lib::sout(file)+".");
  }
}
//////////////////////////////////////////////////////////////////////////////
void region_writeScene(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
)
//////////////////////////////////////////////////////////////////////////////
// aArgs[0] : string : file name
// aArgs[1] : boolean : binary or not
// aArgs[2] : string : what in the scene to write. For exa all or scene.
//            all put the scene but may also put light, lightModel and camera. 
// aArgs[3] : boolean : generate alternate representations.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aArgs.size()!=4) return;
  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;
  SoRegion* soRegion = soPage->currentRegion();
  if(!soRegion) return;

  std::string file = aArgs[0];
  std::string format = Lib::smanip::suffix(file);

  bool binary = false;
  if(!Lib::smanip::tobool(aArgs[1],binary)) binary = false;
  std::string what = aArgs[2];
  bool genAlternateRep = true;
  if(!Lib::smanip::tobool(aArgs[3],genAlternateRep)) genAlternateRep = true;

  SoNode* node = soRegion->getWriteNode(what.c_str());
  if(!node) return;
  if(SoTools::write(*node,
                    file.c_str(),
                    format.c_str(),
                    binary?TRUE:FALSE,
                    genAlternateRep?TRUE:FALSE)==FALSE) {
    aUI.echo("Can't write file "+Lib::sout(file)+".");
  }
}
//////////////////////////////////////////////////////////////////////////////
void region_readInventorFile(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
)
//////////////////////////////////////////////////////////////////////////////
// For backward compatibility.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!aArgs.size()) return;
  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;
  SoRegion* soRegion = soPage->currentRegion();
  if(!soRegion) return;
  std::string file = aArgs[0];
  std::string format = Lib::smanip::suffix(file);
  if(soRegion->readScene(file.c_str(),format.c_str())==FALSE) {
    aUI.echo("Can't read file"+file+".");
  }
}
/////////////////////////////////////////////////////////////////////////////
void region_addFrame(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// aArgs[0] : size
// aArgs[1] : color
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;
  SoRegion* soRegion = soPage->currentRegion();
  if(!soRegion) return;
  if(!soRegion->isOfType(SoDisplayRegion::getClassTypeId())) return;
  SoDisplayRegion* displayRegion = (SoDisplayRegion*)soRegion;

  if(aArgs.size()!=2) return;

  double size;
  if(!Lib::smanip::todouble(aArgs[0],size)) return;

  SoSeparator* stat = displayRegion->getStaticScene();
  if(!stat) return;

  double r,g,b;
  Lib::smanip::torgb(aArgs[1],r,g,b);

  SoSeparator* separator = new SoSeparator();
  separator->setName("sceneGraph");

  SoHighlightMaterial* highlightMaterial = new SoHighlightMaterial;
  highlightMaterial->diffuseColor.setValue
    (SbColor((float)r,(float)g,(float)b));
  //highlightMaterial->transparency.setValue((float)transparency);
  highlightMaterial->highlightColor.setValue(SbColor(1,1,1));
  separator->addChild(highlightMaterial);

  SoCoordinateAxis* frame = new SoCoordinateAxis();
  frame->setName("frame");
  frame->fNDivision.setValue(10);
  frame->fDivisionLength.setValue((float)size);
  separator->addChild(frame);

  stat->addChild(separator);

}
//////////////////////////////////////////////////////////////////////////////
void region_setParameter(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// args[0] what
// args[1,n-1] : partial value
//  value is built with : args[1] + " " + args[2] + ..... + args[n-1]
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;
  SoRegion* soRegion = soPage->currentRegion();
  if(!soRegion) return;
  if(!aArgs.size()) return;
  std::string what = aArgs[0];
  if(aArgs.size()==1) {
    SoTools::setKitPartValue(soRegion,what.c_str(),"");
  } else {
    std::string value = aArgs[1];
    for(unsigned int index=2;index<aArgs.size();index++) 
      value += " " + aArgs[index];
    if(what=="camera") {
      SoCamera* camera = soRegion->getCamera();
      if(camera) {
        if(camera->set(value.c_str())==FALSE) {
          Lib::Out out(aUI.session().printer());
          out << "region_setParameter :"
              << " On camera, unable to set " << Lib::sout(value)
              << Lib::endl;
        }
      }
    } else {
      SoTools::setKitPartValue(soRegion,what.c_str(),value.c_str());
    }
  }
}
//////////////////////////////////////////////////////////////////////////////
void region_toggleParameter(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// args[0] what
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!aArgs.size()) return;
  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;
  SoRegion* soRegion = soPage->currentRegion();
  if(!soRegion) return;
  SoTools::toggleKitPartValue(soRegion,aArgs[0].c_str());
}
//////////////////////////////////////////////////////////////////////////////
void region_resetUndo(
 Slash::UI::IUI& aUI
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;
  SoRegion* soRegion = soPage->currentRegion();
  if(!soRegion) return;
  soRegion->resetUndo();
}
//////////////////////////////////////////////////////////////////////////////
void region_clear(
 Slash::UI::IUI& aUI
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;
  SoRegion* soRegion = soPage->currentRegion();
  if(!soRegion) return;
  soRegion->clear();
}
//////////////////////////////////////////////////////////////////////////////
void region_undo(
 Slash::UI::IUI& aUI
,const std::vector<std::string>&
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;
  SoRegion* soRegion = soPage->currentRegion();
  if(!soRegion) return;
  soRegion->undo();
}
//////////////////////////////////////////////////////////////////////////////
void region_reset_highlight( //FIXME : inconsistent casing
 Slash::UI::IUI& aUI
,const std::vector<std::string>&
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;
  SoRegion* soRegion = soPage->currentRegion();
  if(!soRegion) return;
 //FIXME page->resetCallbackValue();
  SoTools::resetHighlight(soRegion);
  //FIXME page->executeCollectCallback();
  //FIXME page->resetCallbackValue();
}
//////////////////////////////////////////////////////////////////////////////
void region_setSize(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// aArgs[0] : A4H, A4V
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!aArgs.size()) return;
  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;
  SoRegion* soRegion = soPage->currentRegion();
  if(!soRegion) return;
  SoViewportRegion* vpr = soRegion->getViewportRegion();
  unsigned int w,h;
  vpr->getSizePixels(w,h);
  if(!w || !h) return;
  if(aArgs[0]=="A4V") {
    if(w>=h) {
      w = (unsigned int)(float(h)*21.0F/29.7F);
      vpr->setSizePixels(w,h);
    } else {
      h = (unsigned int)(float(w)*29.7F/21.0F);
      vpr->setSizePixels(w,h);
    }
  } else if(aArgs[0]=="A4H") {
    if(w>=h) {
      h = (unsigned int)(float(w)*21.0F/29.7F);
      vpr->setSizePixels(w,h);
    } else {
      w = (unsigned int)(float(h)*29.7F/21.0F);
      vpr->setSizePixels(w,h);
    }
  }
}
//////////////////////////////////////////////////////////////////////////////
void region_dump_highlighted(
 Slash::UI::IUI& aUI
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;
  SoRegion* soRegion = soPage->currentRegion();
  if(!soRegion) return;

  SoNode* node = soRegion->findNode("part","topSeparator");
  if(!node) {
    Lib::Out out(aUI.session().printer());
    out << "region_exa_pick :" 
        << " part not found." 
        << Lib::endl;
    return;
  }
 
  // Need to search in nodekits (for exa SoDetectorTreeKit) :
  SbBool oldsearch = SoBaseKit::isSearchingChildren();
  SoBaseKit::setSearchingChildren(TRUE);
  SoSearchAction searchAction;
  searchAction.setFind(SoSearchAction::NAME);
  searchAction.setName("sceneGraph");
  searchAction.setInterest(SoSearchAction::ALL);
  searchAction.apply(node);
  SoBaseKit::setSearchingChildren(oldsearch);

  SoPathList& pathList = searchAction.getPaths();
  unsigned int number = pathList.getLength();
  if(!number) return;

  Lib::Out out(aUI.session().printer());
  for(unsigned int index=0;index<number;index++) {
    SoFullPath* path = (SoFullPath*)pathList[index];
    SoNode* node = path->getTail();

    bool highlighted = false;
   {// Search if the SoShape under the "sceneGraph" node is highlighted :
    SoSearchAction searchAction;
    searchAction.setFind(SoSearchAction::TYPE);
    searchAction.setType(SoHighlightMaterial::getClassTypeId());
    searchAction.apply(node);
    SoPath* path = searchAction.getPath();
    if(path) {
      SoHighlightMaterial* material = (SoHighlightMaterial*)path->getTail();
      if(material->getSavedMaterial()) highlighted = true;
    }}

    if(highlighted) {
      // Search for a SoShape node under the "sceneGraph" node :
      SoSearchAction searchAction;
      searchAction.setFind(SoSearchAction::TYPE);
      searchAction.setType(SoShape::getClassTypeId());
      searchAction.apply(node);
      SoPath* path = searchAction.getPath();
      std::string spath = path ? path->getTail()->getName().getString():"nil";
      out << std::string(spath) << Lib::endl;
    }
  }
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////
void region_count(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!aArgs.size()) return;
  SoSeparator* sep = 0;
  if(aArgs[0]=="staticScene") {
    SoPage* soPage = ui_SoPage(aUI);
    if(!soPage) return;
    SoRegion* soRegion = soPage->currentRegion();
    if(!soRegion) return;
    if(!soRegion->isOfType(SoDisplayRegion::getClassTypeId())) return;
    SoDisplayRegion* displayRegion = (SoDisplayRegion*)soRegion;
    sep = (SoSeparator*)displayRegion->getStaticScene();
  } else if(aArgs[0]=="dynamicScene") {
    SoPage* soPage = ui_SoPage(aUI);
    if(!soPage) return;
    SoRegion* soRegion = soPage->currentRegion();
    if(!soRegion) return;
    if(!soRegion->isOfType(SoDisplayRegion::getClassTypeId())) return;
    SoDisplayRegion* displayRegion = (SoDisplayRegion*)soRegion;
    sep = (SoSeparator*)displayRegion->getDynamicScene();
  } else {
    return;
  }

  Counter counter;
  counter.fTriangles = 0;
  counter.fLineSegments = 0;
  counter.fPoints = 0;

  SoCallbackAction callbackAction;
  callbackAction.addTriangleCallback
    (SoShape::getClassTypeId(),CountTrianglesCB,(void*)&counter);
  callbackAction.addLineSegmentCallback
    (SoShape::getClassTypeId(),CountLineSegmentsCB,(void*)&counter);
  callbackAction.addPointCallback
    (SoShape::getClassTypeId(),CountPointsCB,(void*)&counter);
  callbackAction.apply(sep);

  SoCounterAction counterAction;
  counterAction.apply(sep);
  int nodes = counterAction.getCount();

  counterAction.setLookFor(SoCounterAction::TYPE);
  counterAction.setType(SoShape::getClassTypeId());
  counterAction.apply(sep);
  int shapes = counterAction.getCount();

  Lib::Out out(aUI.session().printer());
  out << "Number of triangles : " << counter.fTriangles << Lib::endl;
  out << "Number of line segments : " << counter.fLineSegments << Lib::endl;
  out << "Number of points : " << counter.fPoints << Lib::endl;
  out << "Number of nodes : " << nodes << Lib::endl;
  out << "Number of shapes : " << shapes << Lib::endl;
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
/// OnX/scripts/OnX/Pages.onx ////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
void OnX_Pages_Banner(
 Slash::UI::IUI& aUI
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Arguments args(aUI.session().arguments());
  bool b;
  if(args.find("-banner",b)&&(!b)) return; //For exa from opy.

  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;

  soPage->titleVisible.setValue(FALSE);
  soPage->createRegions("SoDisplayRegion",1,1,0);

  SoRegion* soRegion = soPage->currentRegion();
  if(!soRegion) return;

  soRegion->color.setValue(0.8F,0.8F,0.8F);

  std::string file = Lib::fmanip::name("$ONXROOT/scripts/Inventor/OnX.iv");
  std::string format = Lib::smanip::suffix(file);

  if(soRegion->readScene(file.c_str(),format.c_str())==FALSE) {
    aUI.echo("Can't read file"+file+".");
  }
}
//////////////////////////////////////////////////////////////////////////////
void region_addSoNode(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// aArgs[0] node class (def is SoCube)
// aArgs[1] staticScene or dynamicScene (def is staticScene)
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;
  SoRegion* soRegion = soPage->currentRegion();
  if(!soRegion) return;
  Lib::Out out(aUI.session().printer());
  if(!soRegion->isOfType(SoDisplayRegion::getClassTypeId())) {
    out << "region_addSoNode :" 
        << " region not an SoDisplayRegion." 
        << Lib::endl;
    return;
  }
  SoDisplayRegion* displayRegion = (SoDisplayRegion*)soRegion;

  std::string sclass = "SoCube";
  if(aArgs.size()) sclass = aArgs[0];

  SoSeparator* sep = displayRegion->getStaticScene();
  if((aArgs.size()==2) && (aArgs[0]=="dynamicScene")) {
    sep = displayRegion->getDynamicScene();
  }

  SoType soType = SoType::fromName(sclass.c_str());
  if(soType==SoType::badType()) {
    out << "region_addSoNode :" 
        << " unknown class " << Lib::sout(sclass)
        << Lib::endl;
    return;
  }
  if(soType.canCreateInstance()==FALSE) {
    out << "region_addSoNode :" 
        << " can't create instance for class " << Lib::sout(sclass)
        << Lib::endl;
    return;
  }

  SoNode* node = (SoNode*)soType.createInstance();
  if(!node) {
    out << "region_addSoNode :" 
        << " create instance failed for class " << Lib::sout(sclass)
        << Lib::endl;
    return;
  }

  sep->addChild(node);
}
/////////////////////////////////////////////////////////////////////////////
void OnX_InputFrame_ok(
 Slash::UI::IUI& aUI
) 
//////////////////////////////////////////////////////////////////////////////
// input fields :
//  real size
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;
  SoRegion* soRegion = soPage->currentRegion();
  if(!soRegion) return;
  SoSeparator* stat = (SoSeparator*)soRegion->findNode("part","topSeparator");
  if(!stat) return;

  Lib::Out out(aUI.session().printer());

  double size;
  if(!ui_parameterValue(aUI,"OnX_InputFrame_input_size.value",size)) return;

  // Solve environment variables :
  std::string name = Lib::fmanip::name("$ONXROOT/scripts/Inventor/frame.iv");
  SoInput soInput;
  if(!soInput.openFile(name.c_str())) {
    out << "Can't open file " << Lib::sout(name) << Lib::dot_endl;
    return;
  }
  SoSeparator* soFile = SoDB::readAll(&soInput);
  if(!soFile) {
    out << "Can't read file " << Lib::sout(name) << Lib::dot_endl;
    return;
  }

  SoSeparator* separator = new SoSeparator();
  separator->setName("sceneGraph");

  SoScale* scale = new SoScale();
  scale->scaleFactor.setValue(SbVec3f((float)size,(float)size,(float)size));

  separator->addChild(scale);
  separator->addChild(soFile);

  stat->addChild(separator);
}

//////////////////////////////////////////////////////////////////////////////
/// OnX/scripts/OnX/Examples.onx /////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
void OnX_Examples_Shapes(
 Slash::UI::IUI& aUI
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Slash::UI::ISoViewer* soViewer = ui_SoViewer(aUI);
  if(!soViewer) return;
  soViewer->setFeedbackVisibility(true);
  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;

  soPage->title.setValue("OnX/Shapes");

  SoRegion* soRegion = soPage->currentRegion();
  if(!soRegion) return;

  soRegion->color.setValue(0.6F,0.6F,0.6F);
  soRegion->resetUndo();

  Slash::Core::ISession& session = aUI.session();

  session_setColor(session,"red");
  session.setParameter("modeling.modeling","solid");
  session.setParameter("modeling.transparency","0.4");
  data_collect(session,"OnX_Shape","type==\"Cube\"");
  data_visualize(session);

  session_setColor(session,"magenta");
  session.setParameter("modeling.modeling","wire_frame");
  session.setParameter("modeling.transparency","0");
  data_collect(session,"OnX_Shape","(type==\"Hole\")&&(size<=1)");
  data_visualize(session);

  session_setColor(session,"yellow");
  session.setParameter("modeling.modeling","none"); //From style file.
  session.setParameter("modeling.transparency","0");
  data_collect(session,"OnX_Shape","(type==\"Hole\")&&(size>1)");
  data_visualize(session);

  session_setColor(session,"blue");
  session.setParameter("modeling.modeling","texts");
  data_collect(session,"OnX_Shape","(type==\"Sphere\")&&(size<=2)");
  data_visualize(session);

  session_setColor(session,"none");
  session.setParameter("modeling.modeling","none"); //From Shape.styles
  data_collect(session,"OnX_Shape","(type==\"Sphere\")&&(size>2)");
  data_visualize(session,"OnX_Shape_Sphere_big");

  // For Track[1,2], take OnX/scripts/Styles/Shape.styles styles.
  session_modeling_none(session);
  data_collect(session,"OnX_Shape","type==\"Track*\"");
  data_visualize(session);

}
//////////////////////////////////////////////////////////////////////////////
void OnX_Examples_Detector(
 Slash::UI::IUI& aUI
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;
  soPage->title.setValue("OnX/Detector");
  SoRegion* soRegion = soPage->currentRegion();
  if(!soRegion) return;
  if(!soRegion->isOfType(SoDisplayRegion::getClassTypeId())) return;
  SoDisplayRegion* displayRegion = (SoDisplayRegion*)soRegion;
  SoSeparator* sep = (SoSeparator*)displayRegion->getStaticScene();
  if(!sep) return;

  SoDetectorExample* det = new SoDetectorExample();
  sep->addChild(det);
}
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
void OnX_InputPage_ok(
 Slash::UI::IUI& aUI
) 
//////////////////////////////////////////////////////////////////////////////
// input fields :
//  string color
//  string griding
//  string class
//  int connect 
//  string action
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string action;
  if(!aUI.parameterValue("OnX_InputPage_input_action.value",action)) return;

  if(action=="Set color") {

    std::string color;
    if(!aUI.parameterValue("OnX_InputPage_input_color.value",color)) return;

    double r,g,b;
    if(!Lib::smanip::torgb(color,r,g,b)) return;

    SoPage* soPage = ui_SoPage(aUI);
    if(!soPage) return;
    SoRegion* soRegion = soPage->currentRegion();
    if(!soRegion) return;

    soRegion->color.setValue((float)r,(float)g,(float)b);

  } else if(action=="Set grid") {

    std::string griding;
    if(!aUI.parameterValue("OnX_InputPage_input_griding.value",griding)) return;
    std::vector<std::string> words;
    Lib::smanip::words(griding,"x",words);
    if(words.size()!=2) return;

    std::string sclass;
    if(!aUI.parameterValue("OnX_InputPage_input_class.value",sclass)) return;

    int cols;
    if(!Lib::smanip::toint(words[0],cols)) return;

    int rows;
    if(!Lib::smanip::toint(words[1],rows)) return;

    SoPage* soPage = ui_SoPage(aUI);
    if(!soPage) return;

    soPage->addRegions(sclass.c_str(),cols,rows);

  } else if(action=="Connect region") {
    std::string connect;
    if(connect=="") connect = "0";
    if(!aUI.parameterValue("OnX_InputPage_input_connect.value",connect)) return;
    int index;
    if(!Lib::smanip::toint(connect,index)) return;
    
    SoPage* soPage = ui_SoPage(aUI);
    if(!soPage) return;

    soPage->connectCurrentRegion(index);
  }
}
//////////////////////////////////////////////////////////////////////////////
void OnX_InputRegion_ok(
 Slash::UI::IUI& aUI
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string svalue;

  std::string sclass;
  if(!aUI.parameterValue("OnX_InputRegion_input_class.value",sclass)) return;

  double x;
  if(!ui_parameterValue(aUI,"OnX_InputRegion_input_x.value",x)) return;
  double y;
  if(!ui_parameterValue(aUI,"OnX_InputRegion_input_y.value",y)) return;
  double w;
  if(!ui_parameterValue(aUI,"OnX_InputRegion_input_w.value",w)) return;
  double h;
  if(!ui_parameterValue(aUI,"OnX_InputRegion_input_h.value",h)) return;

  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;

  SoRegion* soRegion = 
    soPage->createRegion(sclass.c_str(),(float)x,(float)y,(float)w,(float)h);
  if(!soRegion) return;
  soPage->setCurrentRegion(soRegion);
}

//////////////////////////////////////////////////////////////////////////////
/// Plotter //////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
void plotter_rebuild(
 Slash::UI::IUI& aUI
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;
  SoRegion* soRegion = soPage->currentRegion();
  if(!soRegion) return;
  if(!soRegion->isOfType(SoPlotterRegion::getClassTypeId())) return;
  SoPlotterRegion* soPlotterRegion = (SoPlotterRegion*)soRegion;
  SoPlotter* soPlotter = soPlotterRegion->getPlotter();
  soPlotter->raiseAllUpdateFlags();
  soPlotter->updateChildren();
}
//////////////////////////////////////////////////////////////////////////////
void plotter_set_tickModeling(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!aArgs.size()) return;
  SbString model(aArgs[0].c_str());

  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;
  SoRegion* soRegion = soPage->currentRegion();
  if(!soRegion) return;
  if(!soRegion->isOfType(SoPlotterRegion::getClassTypeId())) return;
  SoPlotterRegion* soPlotterRegion = (SoPlotterRegion*)soRegion;
  SoPlotter* soPlotter = soPlotterRegion->getPlotter();

  soPlotter->setAxesModeling(model);
}
//////////////////////////////////////////////////////////////////////////////
void plotter_set_style(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!aArgs.size()) return;
  std::string style = aArgs[0];

  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;
  SoRegion* soRegion = soPage->currentRegion();
  if(!soRegion) return;
  if(!soRegion->isOfType(SoPlotterRegion::getClassTypeId())) return;
  SoPlotterRegion* soPlotterRegion = (SoPlotterRegion*)soRegion;
  //SoPlotter* plotter = soPlotterRegion->getPlotter();

  soPlotterRegion->setStyle(style.c_str());
}
//////////////////////////////////////////////////////////////////////////////
void plotter_erase(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!aArgs.size()) return;
  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;
  SoRegion* soRegion = soPage->currentRegion();
  if(!soRegion) return;
  if(!soRegion->isOfType(SoPlotterRegion::getClassTypeId())) return;
  SoPlotterRegion* soPlotterRegion = (SoPlotterRegion*)soRegion;
  SoPlotter* soPlotter = soPlotterRegion->getPlotter();
  soPlotter->removePlottableObjectsOfClass(aArgs[0].c_str());  
}
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
void OnX_ImageRegion_ok(
 Slash::UI::IUI& aUI
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Out out(aUI.session().printer());

  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;
  SoRegion* soRegion = soPage->currentRegion();
  if(!soRegion) return;
  if(!soRegion->isOfType(SoImageRegion::getClassTypeId())) {
    out << "OnX_ImageRegion_ok :"  
        << " region not a SoImageRegion." << Lib::endl;
    return;
  }
  SoImageRegion* soImageRegion = (SoImageRegion*)soRegion;
  HEPVis_SoImage* soImage = soImageRegion->getImage();

  std::string fileName;
  if(!aUI.parameterValue
    ("OnX_ImageRegion_input_fileName.value",fileName)) return;
  std::string format;
  if(!aUI.parameterValue("OnX_ImageRegion_input_format.value",format)) return;

  fileName = Lib::fmanip::name(fileName);

  if(format=="guessed") {
    format = Lib::smanip::suffix(fileName);
    Lib::smanip::tolowercase(format);  
  } 

  if(format=="gif")  {
    soImage->format.setValue(HEPVis_SoImage::GIF);
    soImage->fileName.setValue(fileName.c_str());
  } else if( (format=="jpg") || (format=="jpeg"))   {
    soImage->format.setValue(HEPVis_SoImage::JPEG);
    soImage->fileName.setValue(fileName.c_str());
  } else if( (format=="fits") || (format=="fit")) {
    soImage->format.setValue(HEPVis_SoImage::FITS);
    soImage->fileName.setValue(fileName.c_str());
  } else {
    out << "OnX_ImageRegion_ok :"  
        << " unknow format " << Lib::sout(format) << Lib::endl;
  }
}
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
void OnX_TextRegion_ok(
 Slash::UI::IUI& aUI
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Out out(aUI.session().printer());

  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;
  SoRegion* soRegion = soPage->currentRegion();
  if(!soRegion) return;
  if(!soRegion->isOfType(SoTextRegion::getClassTypeId())) {
    out << "OnX_TextRegion_ok :"  
        << " region not a SoTextRegion." << Lib::endl;
    return;
  }
  SoTextRegion* soTextRegion = (SoTextRegion*)soRegion;

  std::string text;
  if(!aUI.parameterValue("OnX_TextRegion_input_text.value",text)) return;
  std::string encoding;
  if(!aUI.parameterValue
    ("OnX_TextRegion_input_encoding.value",encoding)) return;

  soTextRegion->text.setValue(text.c_str());

  SoTextStyle* soStyle = soTextRegion->getStyle(0);
  if(soStyle) {
    soStyle->encoding.setValue(encoding.c_str());
    //soStyle->color.setValue(0,0,1);
    //soStyle->fontName.setValue("times");
  }
}
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
void OnX_region_readScene(
 Slash::UI::IUI& aUI
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  aUI.hide("fileChooser");
  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;
  SoRegion* soRegion = soPage->currentRegion();
  if(!soRegion) return;

  std::string file = aUI.callbackValue();
  std::string format = Lib::smanip::suffix(file);

  if(soRegion->readScene(file.c_str(),format.c_str())==FALSE) {
    aUI.echo("Can't read file"+file+".");
  }
}
//////////////////////////////////////////////////////////////////////////////
void OnX_InputWrite_ok(
 Slash::UI::IUI& aUI
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string file;
  if(!aUI.parameterValue("OnX_InputWrite_input_file.value",file)) return;

  std::string format;
  if(!aUI.parameterValue("OnX_InputWrite_input_format.value",format)) return;

  bool binary = false;
  //if(!ui_parameterValue
  //     (aUI,"OnX_InputWrite_input_binary.value",binary)) return;

  std::string what;
  if(!aUI.parameterValue("OnX_InputWrite_input_what.value",what)) return;

  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;

  if(format=="from_extension") {
    format = Lib::smanip::suffix(file);
  }

  if(what=="page") {
    if(SoTools::write(*soPage,
                      file.c_str(),
                      format.c_str(),
                      binary?TRUE:FALSE,FALSE)==FALSE) {
      aUI.echo("Can't write file "+Lib::sout(file)+".");
    }
  } else if(what=="region") {
    SoRegion* soRegion = soPage->currentRegion();
    if(!soRegion) return;
    if(SoTools::write(*soRegion,
                      file.c_str(),
                      format.c_str(),
                      binary?TRUE:FALSE,FALSE)==FALSE) {
      aUI.echo("Can't write file "+Lib::sout(file)+".");
    }
  } else { 
    SoRegion* soRegion = soPage->currentRegion();
    if(!soRegion) return;
    SoNode* node = soRegion->getWriteNode(what.c_str());
    if(!node) return;
    if(SoTools::write(*node,
                      file.c_str(),
                      format.c_str(),
                      binary?TRUE:FALSE,FALSE)==FALSE) {
      aUI.echo("Can't write file "+Lib::sout(file)+".");
    }
  }
}
//////////////////////////////////////////////////////////////////////////////
/// ruler ////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
void OnX_ruler(
 Slash::UI::IUI& aUI
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Slash::UI::ISoViewer* soViewer = ui_SoViewer(aUI);
  if(!soViewer) return;
  soViewer->setBackgroundColor(1,1,1);
  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;

  soPage->deleteRegions();
  soPage->titleVisible.setValue(FALSE);

  float w = 0.9F;
  float h = 0.9F;

  // Visualization region :
  SoDisplayRegion* displayRegion = 
    (SoDisplayRegion*)soPage->createRegion("SoDisplayRegion",
                                           1-w,1-h,w,h);
  if(!displayRegion) return;
  soPage->setCurrentRegion(displayRegion);
  displayRegion->getViewportRegion()->backgroundColor.setValue(SbColor(0,0,0));

  // Horizontal ruler region :
  SoRulerRegion* hRuler = 
    (SoRulerRegion*)soPage->createRegion("SoRulerRegion",
                                         1-w,0,w,1-h);
  if(!hRuler) return;
  hRuler->getViewportRegion()->backgroundColor.setValue(SbColor(1,1,1));

  // Connect the display region to the ruler :
  hRuler->setMeasuredRegion(displayRegion);

  // Vertical ruler region :
  SoRulerRegion* vRuler = 
    (SoRulerRegion*)soPage->createRegion("SoRulerRegion",
                                         0,1-h,1-w,h);
  if(!vRuler) return;
  vRuler->getViewportRegion()->backgroundColor.setValue(SbColor(1,1,1));

  // Connect the display region to the ruler :
  vRuler->setMeasuredRegion(displayRegion);

  // Set display region as current :
  soPage->setCurrentRegion(0);

}
//////////////////////////////////////////////////////////////////////////////
void OnX_InputDetectIntersection_ok(
 Slash::UI::IUI& aUI
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string what;
  if(!aUI.parameterValue
    ("OnX_InputDetectIntersection_input_what.value",what)) return;

  std::string show;
  if(!aUI.parameterValue
    ("OnX_InputDetectIntersection_input_show.value",show)) return;

  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;
  SoRegion* soRegion = soPage->currentRegion();
  if(!soRegion) return;
  if(!soRegion->isOfType(SoDisplayRegion::getClassTypeId())) return;
  SoDisplayRegion* displayRegion = (SoDisplayRegion*)soRegion;

  SoSeparator* sep = 0;
  if(what=="scene"){
    sep = (SoSeparator*)displayRegion->getSceneSeparator();
  } else if(what=="staticScene"){
    sep = (SoSeparator*)displayRegion->getStaticScene();
  } else if(what=="dynamicScene") {
    sep = (SoSeparator*)displayRegion->getDynamicScene();
  }

  if(!sep) return;

  SoTools::removeManips(*sep);

  if(show=="first") {
    SoTools::detectIntersection(*sep,TRUE);
  } else {
    SoTools::detectIntersection(*sep,FALSE);
  }
}
//////////////////////////////////////////////////////////////////////////////
void OnX_removeManips(
 Slash::UI::IUI& aUI
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  SoPage* soPage = ui_SoPage(aUI);
  if(!soPage) return;
  SoRegion* soRegion = soPage->currentRegion();
  if(!soRegion) return;
  if(!soRegion->isOfType(SoDisplayRegion::getClassTypeId())) return;
  SoDisplayRegion* displayRegion = (SoDisplayRegion*)soRegion;
  SoSeparator* sep = (SoSeparator*)displayRegion->getSceneSeparator();
  SoTools::removeManips(*sep);
}

//////////////////////////////////////////////////////////////////////////////
void OnX_exa_viewer(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aArgs.size()!=1) return;

  std::string name = aUI.callbackData().component();

  Slash::UI::IWidget* widget = aUI.findWidget(name);
  if(!widget) return;
  //aUI.echo("OnX_exa_viewer : "+name);
  Slash::UI::ISoViewer* soViewer = Lib_Cast(*widget,Slash::UI::ISoViewer);
  if(!soViewer) return;

  SoSeparator* sep = (SoSeparator*)soViewer->sceneGraph();
  if(!sep) return;

  Lib::Out out(aUI.session().printer());

  // Solve environment variables :
  std::string file = Lib::fmanip::name(aArgs[0]);
  SoInput soInput;
  if(!soInput.openFile(file.c_str())) {
    out << "Can't open file " << Lib::sout(file) << Lib::dot_endl;
    return;
  }
  SoSeparator* soFile = SoDB::readAll(&soInput);
  if(!soFile) {
    out << "Can't read file " << Lib::sout(file) << Lib::dot_endl;
    return;
  }

  sep->addChild(soFile);
}

} // extern "C"
