#include <Slash/Core/ISession.h>
#include <Slash/Core/IManager.h>
#include <Lib/Out.h>
#include <Lib/Manager.h>
#include <Lib/smanip.h>

#include <Slash/UI/IScriptManager.h>
#include <Slash/UI/IInterpreter.h>

extern "C" {
  void OnXJavaInitialize(Slash::Core::ISession&);
  void OnXJavaFinalize(Slash::Core::ISession&);
  Slash::Core::ISession* OnXJavaGetSession();
}

#include <stdlib.h>

#include <jni.h>

namespace OnX {
class Java {
public:
  Java(Slash::Core::ISession& aSession)
  :fSession(aSession)
  ,fJavaVM(0)    
  ,fJNIEnv(0)
  ,fOnXScript(0)
  ,fExecMethod(0)
  ,fExecClassMethod(0)
  {
    if(!getenv("CLASSPATH")) {
      Lib::Out out(fSession.printer());
      out << "OnX::Java::Java :"
          << " env variable CLASSPATH not setted." << Lib::endl;
      return;    
    }
    //fSession.out().println("debug : CLASSPATH :%s",getenv("CLASSPATH"));
    std::string opt;
    opt += "-Djava.class.path=";
    opt += getenv("CLASSPATH");
    JavaVMOption options[2];
    options[0].optionString = (char*)opt.c_str();
    options[1].optionString = (char*)"-verbose";
    JavaVMInitArgs args;
    args.version = JNI_VERSION_1_2;
    args.options = options;
    args.nOptions = 1;
    args.ignoreUnrecognized = 1;
    int rc = JNI_CreateJavaVM(&fJavaVM,(void **)&fJNIEnv,&args);
    if(rc < 0) {
      Lib::Out out(fSession.printer());
      out << "OnX::Java::Java :"
          << " Failed to create Java VM." << Lib::endl;
      fJavaVM = 0;
      fJNIEnv = 0;
      return;
    }
    // Build an OnX/Script object :
    jclass cls = fJNIEnv->FindClass("OnX/Script");
    if(!cls) {
      Lib::Out out(fSession.printer());
      out << "OnX::Java::Java :"
          << " Could not find OnX/Script class" << Lib::endl;
      return;
    }
    //fSession.out().println("class OnX/Script found");
    jmethodID constructor = fJNIEnv->GetMethodID(cls,"<init>","()V");
    if(!constructor) {
      Lib::Out out(fSession.printer());
      out << "OnX::Java::Java :"
          << " Could not find OnX/Script constructor." << Lib::endl;
      return;
    }
    fExecMethod = 
      fJNIEnv->GetMethodID(cls,"execute","(Ljava/lang/String;J)V");
    if(!fExecMethod) {
      Lib::Out out(fSession.printer());
      out << "OnX::Java::Java :"
          << " Could not find OnX/Script/execute method." << Lib::endl;
      return;
    }
    fExecClassMethod = 
      fJNIEnv->GetMethodID(cls,"executeClass","(Ljava/lang/String;J)Z");
    if(!fExecClassMethod) {
      Lib::Out out(fSession.printer());
      out << "OnX::Java::Java :"
          << " Could not find OnX/Script/executeClass method." << Lib::endl;
      return;
    }
    fOnXScript = fJNIEnv->NewObject(cls,constructor);
    if(!fOnXScript) {
      Lib::Out out(fSession.printer());
      out << "OnX::Java::Java :"
          << " Could not create an OnX/Script object" << Lib::endl;
      return;
    }
  }
  ~Java(){
    if(fJNIEnv && fOnXScript) fJNIEnv->DeleteLocalRef(fOnXScript);
    if(fJavaVM) fJavaVM->DestroyJavaVM();
  }
  void execute(const std::string& aString) {
    if(!fJNIEnv) return;
    if(!fOnXScript) return;
    jstring jstr = fJNIEnv->NewStringUTF(aString.c_str());
    //jlong sess = (jlong)&fSession;
    jlong sess = *((jlong*)((void*)&fSession)); //fix warning on Linux-32 bits
    fJNIEnv->CallVoidMethod(fOnXScript,fExecMethod,jstr,sess);
  }
  void executeClass(const std::string& aName) {
    if(!fJNIEnv) return;
    if(!fOnXScript) return;
    jstring jstr = fJNIEnv->NewStringUTF(aName.c_str());
    //jlong sess = (jlong)&fSession;
    jlong sess = *((jlong*)((void*)&fSession)); //fix warning on Linux-32 bits
    fJNIEnv->CallBooleanMethod(fOnXScript,fExecClassMethod,jstr,sess);
  }
private:
  Slash::Core::ISession& fSession;
  JavaVM* fJavaVM;       
  JNIEnv* fJNIEnv;
  jobject fOnXScript;
  jmethodID fExecMethod;
  jmethodID fExecClassMethod;
};
}

static bool java_Function(const std::string&,void*);

static Slash::Core::ISession* sSession = 0; // Beurk. 
//////////////////////////////////////////////////////////////////////////////
Slash::Core::ISession* OnXJavaGetSession(
) 
//////////////////////////////////////////////////////////////////////////////
// So that we can find the session from a Java script
// but also work in a SWIG context.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return sSession;
}
//////////////////////////////////////////////////////////////////////////////
void OnXJavaInitialize(
 Slash::Core::ISession& aSession
) 
//////////////////////////////////////////////////////////////////////////////
// Executed, by Interpreters, at first java> java code.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  sSession = &aSession; // Rebeurk.

  Slash::UI::IScriptManager* scriptManager = 
    Lib_findManager(aSession,"ScriptManager",Slash::UI::IScriptManager);
  if(scriptManager) {
    Slash::UI::IInterpreter* interp = scriptManager->findInterpreter("java");
    if(interp) {
      OnX::Java* java = new OnX::Java(aSession);
      interp->setFunction(java_Function,java);
      java->executeClass("OnX.Initialize");
    }
  }
}
//////////////////////////////////////////////////////////////////////////////
void OnXJavaFinalize(
 Slash::Core::ISession& aSession
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Slash::UI::IScriptManager* scriptManager = 
    Lib_findManager(aSession,"ScriptManager",Slash::UI::IScriptManager);
  if(scriptManager) {
    Slash::UI::IInterpreter* interp = scriptManager->findInterpreter("java");
    if(interp) {
      void* tag;
      interp->function(tag);
      if(tag) delete ((OnX::Java*)tag);
    }
  }
  sSession = 0;
}
//////////////////////////////////////////////////////////////////////////////
bool java_Function(
 const std::string& aString
,void* aTag
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  OnX::Java* onxJava = (OnX::Java*)aTag;
  if(!onxJava) return false;
  std::string cmd = aString;
  Lib::smanip::strip(cmd);
  Lib::smanip::replace(cmd,"\\n\\","");
  if( (cmd.length()>=2) && (cmd[0]=='x') & (cmd[1]==' ') ) {
    std::string cmd2 = cmd.substr(2,cmd.length()-2);
    Lib::smanip::strip(cmd2);
    onxJava->executeClass(cmd2);
  } else {
    onxJava->execute(cmd);
  }
  return true;
}




