#include <Slash/Core/ISession.h>
#include <Slash/Core/IManager.h>
#include <Slash/Core/IWriter.h>

#include <Slash/UI/IScriptManager.h>
#include <Slash/UI/IInterpreter.h>

#include <Lib/smanip.h>
#include <Lib/cstring.h>
#include <Lib/Manager.h>
#include <Lib/Cast.h>
#include <Lib/Out.h>
#include <Lib/sout.h>

extern "C" {
void OnXPythonInitialize(Slash::Core::ISession&);
void OnXPythonFinalize(Slash::Core::ISession&);
}

#if defined(__linux)
// Clash between os_defines.h (coming from <string>) and pyconfig.h
#if defined(_POSIX_C_SOURCE)
#undef _POSIX_C_SOURCE
#endif
#endif

#include <Python.h>

// WARNING : take care that the below functions be no more
//  used when the DLL will be closed.
static bool python_Function(const std::string&,void*);
extern "C" {
static PyObject* sessionPointer(PyObject*,PyObject*);
}

namespace OnX {

class PythonManager : public virtual Slash::Core::IManager {
public: //Slash::Core::IManager
  virtual std::string name() const { return fName;}
  virtual void* cast(const std::string& aClass) const {
         if_Lib_SCast(OnX::PythonManager)
    else if_Lib_SCast(Slash::Core::IManager)
    else return 0;
  }
public:
  PythonManager(Slash::Core::ISession& aSession)
  :fName("PythonManager")
  ,fSession(aSession)
  ,fInitedByOnX(false)
  ,fMethods(0)
  ,fModuleTag(0)
  ,fModule(0)
  //,fThreadState(0)
  {
    // We have to declare the OnX session to Python.
    if(!Py_IsInitialized()) {
      //FIXME : should pass a copied string. 
      // But in fact, do we want to complicate life in passing something ?
      // If passing something, have to restore something at end.
      //Py_SetProgramName((char*)"OnX::PythonManager");
      Py_Initialize();
      fInitedByOnX =  true;
    }
    PyEval_InitThreads();

    fModuleTag = PyCObject_FromVoidPtr(&fSession,NULL); 
    //fModuleTag->ob_refcnt = 1

    // See Include/modsupport.h Python/modsupport.c :
    //FIXME : module should be suffixed by some Slash::Core::ISession::name().
    std::string moduleName("OnX_PythonManager");     
    // WARNING :
    //  The below return a singleton for the whole Python session.
    // Then passing here twice with the same moduleName will return 
    // the same object without increasing its ob_refcnt.
    //  The second time, the new fModuleTag replaces the previous one
    // which is then dereferenced (the new one being referenced).
    fMethods = new PyMethodDef[2];
    fMethods[0].ml_name = Lib_strdup("sessionPointer");
    fMethods[0].ml_meth = sessionPointer;
    fMethods[0].ml_flags = 0;
    fMethods[0].ml_doc = NULL; 

    fMethods[1].ml_name = NULL;
    fMethods[1].ml_meth = NULL;
    fMethods[1].ml_flags = 0;
    fMethods[1].ml_doc = NULL; 
    fModule = Py_InitModule4((char*)moduleName.c_str(),
                             fMethods, 
			     (char*)NULL,fModuleTag,
                             PYTHON_API_VERSION);
    //We are not the creator of fModule, we don't have to DECREF at end.
    if(fModule==NULL) {
      Lib::Out out(fSession.printer());
      out << "OnX::PythonManager::PythonManager :"
          << " Py_InitModule4 failed for " 
          << Lib::sout(moduleName) << "."
          << Lib::endl;
      return;
    }
    //fModule->ob_refcnt = 1 // the first time. 
                             // But may be more if having done 
                             // a "import <moduleName>"
                             // and passed here again.
    //fModuleTag->ob_refcnt = 2

    //enableThread();
  }
  virtual ~PythonManager() { 
    //disableThread();
    if(fModuleTag) {
      Py_DECREF(fModuleTag); //Since we are the creator.
      fModuleTag = 0;
    }
    if(fModule) {
      PyObject* d = PyModule_GetDict(fModule);
      if(d) {
        PyDict_DelItemString(d,fMethods[0].ml_name);
      }
      //Py_DECREF(fModule); //We are not the creator.
      fModule = 0;
    }
    if(fInitedByOnX) {
      Py_Finalize();
    }
    Lib_strdel(fMethods[0].ml_name);
    delete [] fMethods;
  }
  bool isValid() const {
    //return fPyLoaded?true:false;
    return fModule?true:false;
  }
  //void disableThread() {
  //  if(!fThreadState) return; //done
  //  PyEval_RestoreThread(fThreadState);
  //  fThreadState = 0;
  //}
  //void enableThread() {
  //  if(fThreadState) return; //done
  //  fThreadState = PyEval_SaveThread();
  //}
private:
  std::string fName;
  Slash::Core::ISession& fSession;
  bool fInitedByOnX;
  PyMethodDef* fMethods;
  PyObject* fModuleTag;
  PyObject* fModule;
  //PyThreadState* fThreadState;
};

}

//////////////////////////////////////////////////////////////////////////////
void OnXPythonInitialize(
 Slash::Core::ISession& aSession
) 
//////////////////////////////////////////////////////////////////////////////
// Executed, by Interpreters, at first Python>... command
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  OnX::PythonManager* pythonManager =
    Lib_findManager(aSession,"OnX::PythonManager",OnX::PythonManager);
  if(!pythonManager) {
    pythonManager = new OnX::PythonManager(aSession);
    aSession.addManager(pythonManager);
  }

  Slash::UI::IScriptManager* scriptManager = 
    Lib_findManager(aSession,"ScriptManager",Slash::UI::IScriptManager);
  if(scriptManager) {
    //Interpreter should have been declared in Session.cxx.
    Slash::UI::IInterpreter* interp = scriptManager->findInterpreter("Python");
    if(interp) {
      interp->setFunction(python_Function,pythonManager);
    }
  }
}
//////////////////////////////////////////////////////////////////////////////
void OnXPythonFinalize(
 Slash::Core::ISession& aSession
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  aSession.destroyManager("PythonManager");
}
//////////////////////////////////////////////////////////////////////////////
bool python_Function(
 const std::string& aString
,void* aPythonManager
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  OnX::PythonManager* pythonManager = (OnX::PythonManager*)aPythonManager;
  if(!pythonManager->isValid()) return false;

  //FIXME : script may contain usage of '\n' !
  std::vector<std::string> text;
  Lib::smanip::words(aString,"\\n\\",text);
  text.push_back(""); //Text should finish with an empty ligne.
  // Justify text on the left :
  std::vector<std::string> lines;
  Lib::smanip::justifylines(text,lines);
  std::string s = Lib::smanip::tostring(lines,"\n");

#if (PY_MAJOR_VERSION <= 2) && (PY_MINOR_VERSION <= 2)
#else 
  PyGILState_STATE state = PyGILState_Ensure();
#endif

  PyRun_SimpleString((char*)s.c_str());

#if (PY_MAJOR_VERSION <= 2) && (PY_MINOR_VERSION <= 2)
#else 
  PyGILState_Release(state);
#endif

  return true;
}
//////////////////////////////////////////////////////////////////////////////
PyObject* sessionPointer(
 PyObject* aTag
,PyObject* /*aArgs*/ 
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  //aTag is the PythonManager::fModuleTag object.

  Slash::Core::ISession* session = 
    (Slash::Core::ISession*)(void*)PyCObject_AsVoidPtr(aTag);

  // The format must be the same than the 
  // OnX.i/OnX_cast_ISession function.
  std::string s = Lib::smanip::p2sx(session);

  return PyString_FromString(s.c_str());

}
