
#include <AIDA/IAnalysisFactory.h>
#include <AIDA/ITreeFactory.h>
#include <AIDA/ITree.h>
#include <AIDA/IHistogramFactory.h>
#include <AIDA/IHistogram1D.h>
#include <AIDA/IHistogram2D.h>
#include <AIDA/ITupleFactory.h>
#include <AIDA/ITuple.h>
#include <AIDA/IPlotterFactory.h>
#include <AIDA/IPlotter.h>
#include <AIDA/IPlotterRegion.h>

#include <Lib/Arguments.h>
#include <Lib/Randoms.h>
#include <Lib/Printer.h>
#include <Lib/Debug.h>
#include <Lib/smanip.h>

#include <iostream>

//////////////////////////////////////////////////////////////////////////////
int main(
 int aArgc
,char** aArgv
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{

 {//  Put everything in a block so that temporary objects
  // on the stack are deleted before the final memory decount is done. 

  Lib::Arguments args(aArgc,aArgv);
  bool help = args.isAnArgument("-h");
  if(help) {
    std::cout << "Options : " << std::endl;
    std::cout << "  OS> OnXLabTemplateOne_batch -h" << std::endl;
    std::cout << "   to have help." << std::endl;
    std::cout << "  OS> OnXLabTemplateOne_batch -plotting" << std::endl;
    std::cout << "   to run without the plotting (batch mode)." << std::endl;
    std::cout << "  OS> OnXLabTemplateOne_batch -store=[memory,xml,root]" << std::endl;
    std::cout << "   to store :"
              << "     in memory"
              << "     in an xml file"
              << "     in a file at the ROOT format with OnXLab objects."
              << std::endl;
    std::cout << "  OS> OnXLabTemplateOne_batch -store=root -export=root" << std::endl;
    std::cout << "   to store :"
              << "     in a file at the ROOT format with ROOT objects."
              << std::endl;
    return 0;
  }

  std::string store;
  if(!args.find("-store",store)) store = "root";
  Lib::smanip::tolowercase(store);
  bool compress  = args.isAnArgument("-compress");
  std::string sexport;
  if(!args.find("-export",sexport)) sexport = "OnXLab";

  AIDA::IAnalysisFactory* analysisFactory = AIDA_createAnalysisFactory();
  if(!analysisFactory) {
    std::cout << "Can't get an IAnalysisFactory." << std::endl;
    return 1;
  }

  AIDA::ITreeFactory* treeFactory = analysisFactory->createTreeFactory();
  if(!treeFactory) {
    std::cout << "Can't get an ITreeFactory." << std::endl;
    return 1;
  }

  int count;

  // Create some random engines :
  Lib::RandomGauss randomGauss;
  Lib::RandomFlat randomFlat;
  Lib::RandomBreitWigner randomBreitWigner;

  AIDA::ITree* tree = 0;
  std::string storeType;
  std::string storeName;
  if(store=="memory") {
    storeType = "memory";
    storeName = "memory";
    tree = treeFactory->create();
  } else {
    std::string option;
    option += compress ? "compress=yes" : "compress=no";
    storeName = "OnXLabTemplateOne";
    if(store=="root") {
      storeType = "root";
      storeName += ".root";
      option += std::string(";export=")+sexport;
    } else if(store=="xml") {
      storeType = "xml";
      storeName += ".aida";
    }
    tree = treeFactory->create(storeName,storeType,false,true,option);
  }
  if(!tree) {
    std::cout << "Can't get an ITree." << std::endl;
    return 1;
  }
  tree->mkdir("histograms");
  tree->cd("histograms");

  // Get an histogram factory :
  AIDA::IHistogramFactory* histoFactory = 
    analysisFactory->createHistogramFactory(*tree);

  // Create some histos :
  AIDA::IHistogram1D* gauss = 
    histoFactory->createHistogram1D("gauss","Random gauss",100,-5,5);
  for(count=0;count<5000;count++) {
    gauss->fill(randomGauss.shoot());
  }

  AIDA::IHistogram1D* flat = 
    histoFactory->createHistogram1D("10","Random flat",100,0,1);
  for(count=0;count<5000;count++) {
    flat->fill(randomFlat.shoot());
  }

  AIDA::IHistogram2D* gauss_BW = 
    histoFactory->createHistogram2D("gauss_BW",
                                    "BW/Gauss",
                                    100,-5,5,
                                    100,-5,5);
  for(count=0;count<1000;count++) {
    gauss_BW->fill(randomGauss.shoot(),randomBreitWigner.shoot());
  }
  AIDA::IHistogram1D* BW = 
    histoFactory->createHistogram1D("BW","Random BW",100,-5,5);
  for(count=0;count<5000;count++) {
    BW->fill(randomBreitWigner.shoot());
  }

  delete histoFactory; // An histo factory is not "managed", we have to 
                       // delete it ourself.

  if(args.isAnArgument("-plotting")) {

    // Plotting :
    AIDA::IPlotterFactory* plotterFactory = 
      analysisFactory->createPlotterFactory(aArgc,aArgv);
    if(plotterFactory) {
      AIDA::IPlotter* plotter = plotterFactory->create();
      if(plotter) {
        plotter->show();
        plotter->clearRegions();
        AIDA::IPlotterRegion& region = plotter->currentRegion();
        region.plot(*BW);
        std::cout << "Click the escape (or exit, quit) button to continue..." 
                  << std::endl;
        // The user have to click in some "escape GUI" button
        // to return from "interact".
        plotter->interact();
  
        // Demonstrate how to update the plotter :
        region.clear();
        region.plot(*gauss_BW);
  
        if(args.isAnArgument("-xyModeling=color")) {
          region.setParameter("xyModeling","COLOR");
        }
        //plotter->setParameter("infosVisible","TRUE");
        // [0 , 10[ = white ; [10, 50[ = yellow ; [50,oo[ red :
        region.setParameter("colorMapping","yellow 10 orange 40 red ");
        for(int update = 0;update<3;update++) {
          // New data :
                for(int count = 0;count<5000;count++) {
                gauss_BW->fill(randomGauss.shoot(),randomBreitWigner.shoot());
          }
          // Update the plotter :
          plotter->refresh();
          std::cout << "Click the escape button to continue ("
                    << update
                    << ")..."
                    << std::endl;
          plotter->interact();
        }
        delete plotter;
      } else {
        std::cout << "Plotter not found." << std::endl;
      }
      delete plotterFactory;
    }
  }

  tree->cd("..");
  tree->mkdir("tuples");
  tree->cd("tuples");

  // Get a tuple factory :
  AIDA::ITupleFactory* tupleFactory = 
    analysisFactory->createTupleFactory(*tree);

  // Create some tuples :
  AIDA::ITuple* tuple1 = tupleFactory->create("tuple1","tuple1",
    "double v1,double v2");
  // Fill so that an histo will give a step function :
  for(count=0;count<5;count++) {
    for(int i=0;i<count;i++) {
      tuple1->fill(0,(double)count);
      tuple1->fill(1,(double)count);
      tuple1->addRow();
    }
  }
  for(count=5;count<10;count++) {
    for(int i=0;i<count;i++) {
      tuple1->fill(0,(double)count);
      tuple1->fill(1,(double)count);
      tuple1->addRow();
    }
  }

  AIDA::ITuple* tuple2 = tupleFactory->create("tuple2","tuple2",
    "double rand_gauss, double rand_BW, double step");
  for(count=0;count<5000;count++) {
    tuple2->fill(0,randomGauss.shoot());
    tuple2->fill(1,randomBreitWigner.shoot());
    tuple2->fill(2,(double)count);
    tuple2->addRow();
  }

  //tree->ls();

  tree->commit(); // Flush the tuple in the file associated to the ITree.

  std::cout << std::string("Produce ")+storeName+"..." << std::endl;
  tree->ls();
  std::cout << "pwd (" << tree->pwd() << ")" << std::endl;

  delete tupleFactory;

  delete treeFactory;
  delete analysisFactory; // Will delete the trees.
  }

  // A simple check to see if object creation/destruction balance is correct :
  Lib::Printer printer;
  Lib::Debug::balance(printer);

  std::cout << "Exiting..." << std::endl;

  return 0;
}

